import collections
from multiprocessing import Pool, cpu_count


class EmbarrassingParallelisation:
    """
    This class provides a context manager for embarrassingly parallel
    environments that can be used by invoking a 'with' statement.

    """

    def __init__(self, number_of_processes, number_of_ops, verbosity=0):
        # Default to using all cores if either too many or "-1" are
        # specified.
        if number_of_processes > cpu_count() or number_of_processes == -1:
            number_of_processes = cpu_count()
        if verbosity >= 3:
            print(
                "Starting parallel processing pool with up"
                " to {} cores.".format(number_of_processes))
        self.number_of_processes = number_of_processes

        self.number_of_ops = number_of_ops
        chunk = collections.namedtuple('chunk', 'start finish')
        chunks = []
        step_size = number_of_ops // number_of_processes
        if step_size > 1:
            begin = -1
            for i in range(number_of_processes):
                end = begin + step_size
                if number_of_ops - step_size < end:
                    end = number_of_ops
                chunks.append(chunk(start=begin + 1, finish=end))
                begin += step_size
        else:
            chunks.append(chunk(start=0, finish=number_of_ops))
            chunks.append(chunk(start=0, finish=number_of_ops))
        self.chunks = chunks
        self.pool = Pool(processes=self.number_of_processes)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.pool.terminate()


if __name__ == '__main__':
    pass



###############################################################################
#                                                                             #
# This is the part of the program where standard print messages are defined   #
#                                                                             #
###############################################################################

import sys
import time


def timer(start, end):
    hours, rem = divmod(end - start, 3600)
    minutes, seconds = divmod(rem, 60)
    string = "{:0>2}h {:0>2}m {:05.2f}s".format(int(hours), int(minutes),
                                                seconds)
    return string


def msg_timestamp(s, starttime=None):
    s = s + time.strftime("%Y/%m/%d %X")
    if starttime is not None:
        s = s + "\nTime elapsed {}".format(timer(starttime, time.time()))
    return s


def msg_program_header(module: str, version: float):
    # ASCII FONTS from: http://patorjk.com/software/taag/
    # Font = "Big"
    print("###################################################")
    print(" Wellington Fast Assessment of Reactions: WellFaRe")
    print(" Module: {}".format(module))
    print("  __          __  _ _ ______      _____           ")
    print("  \ \        / / | | |  ____/\   |  __ \          ")
    print("   \ \  /\  / /__| | | |__ /  \  | |__) |__       ")
    print("    \ \/  \/ / _ \ | |  __/ /\ \ |  _  // _ \     ")
    print("     \  /\  /  __/ | | | / ____ \| | \ \  __/     ")
    print("      \/  \/ \___|_|_|_|/_/    \_\_|  \_\___|     ")
    print("                                  Version {: .2f} ".format(version))
    print("    WellFARe Copyright (C) 2018 Matthias Lein     ")
    print(" This program comes with ABSOLUTELY NO WARRANTY   ")
    print(" This is free software, and you are welcome to    ")
    print("   redistribute it under certain conditions.      ")
    msg_timestamp('Program started at: ')
    print("###################################################\n")


def msg_program_footer(starttime=None, exit_code=None):
    print("\n###################################################")
    print(msg_timestamp('Program terminated at: '))
    if starttime is not None:
        print("Total time elapsed {}".format(timer(starttime, time.time())))
    print("###################################################")
    if exit_code is not None:
        sys.exit(exit_code)


def msg_program_warning(s):
    print("\n###################################################")
    print("    __          __              _                  ")
    print("    \ \        / /             (_)                 ")
    print("     \ \  /\  / /_ _ _ __ _ __  _ _ __   __ _      ")
    print("      \ \/  \/ / _` | '__| '_ \| | '_ \ / _` |     ")
    print("       \  /\  / (_| | |  | | | | | | | | (_| |     ")
    print("        \/  \/ \__,_|_|  |_| |_|_|_| |_|\__, |     ")
    print("                                         __/ |     ")
    print("                                        |___/      ")
    msg_timestamp('Warning time/date: ')
    print(s)
    print("###################################################\n")
    sys.stdout.flush()
    return


def msg_program_error(s: str = "An error has occurred, exiting now",
                      starttime=None, destination=sys.stdout):
    print(
        "\n###################################################",
        file=destination)
    print("  ______                     ", file=destination)
    print(" |  ____|                    ", file=destination)
    print(" | |__   _ __ _ __ ___  _ __ ", file=destination)
    print(" |  __| | '__| '__/ _ \| '__|", file=destination)
    print(" | |____| |  | | | (_) | |   ", file=destination)
    print(" |______|_|  |_|  \___/|_|   ", file=destination)
    print(msg_timestamp('Error time/date: ', starttime), file=destination)
    print(s, file=destination)
    print(
        "###################################################\n",
        file=destination)
    sys.exit(-1)


def main():
    # Record start time
    prg_start_time = time.time()
    # Print some messages
    msg_program_header("Messages", 1.0)
    # msg_program_warning("Demonstration of a warning message")
    msg_program_error("Demonstration of an error message", prg_start_time)
    # msg_program_error("Error messages can also be sent to stderr",
    #                   prg_start_time,
    #                   destination=sys.stderr)
    msg_program_footer(prg_start_time)


if __name__ == '__main__':
    main()



###############################################################################
#                                                                             #
# This is the part of the program where the methods for reading QM files are  #
# defined                                                                     #
#                                                                             #
###############################################################################

import numpy as np

from typing import Any

from wellfare.constants import *
from wellfare.conversions import *
from wellfare.messages import *
from wellfare.atom import Atom
from wellfare.molecule import Molecule, build_molecular_dihedrals, \
    build_molecular_angles, build_bond_orders


def print_add_atom(symbol: Any, xcoord: Any, ycoord: Any, zcoord: Any):
    """
    This method will print a status message after an atom is being added to the
    current molecule. Note that this method will convert
    coordinates into floats before printing if necessary.

    :param symbol: The atomic symbol of the atom
    :param xcoord: Cartesian x-coordinate.
    :param ycoord: Cartesian y-coordinate.
    :param zcoord: Cartesian z-coordinate.
    :return: None
    """
    print(
        " Adding atom: {:<3} {: 13.8f} {: 13.8f}"
        " {: 13.8f} to molecule.".format(str(symbol), float(xcoord),
                                         float(ycoord), float(zcoord)))


def print_found_atom(symbol: Any, xcoord: Any, ycoord: Any, zcoord: Any):
    """
    This method will print a status message after an atom has been found in the
    input stream while reading a file. Note that this method will convert
    coordinates into floats before printing if necessary.

    :param symbol: The atomic symbol of the atom
    :param xcoord: Cartesian x-coordinate.
    :param ycoord: Cartesian y-coordinate.
    :param zcoord: Cartesian z-coordinate.
    :return: None
    """
    print(
        " Found atom: {:<3} {: 13.8f} {: 13.8f}"
        " {: 13.8f} while reading file.".format(str(symbol), float(xcoord),
                                                float(ycoord), float(zcoord)))


def read_gauss_bond_orders(filename, molecule, verbosity=0):
    bo = []
    f = open(filename, 'r')
    for line in f:
        if line.find(
                "Atomic Valencies and Mayer Atomic Bond Orders:") != -1:
            bo = np.zeros((molecule.num_atoms(), molecule.num_atoms()))
            if verbosity >= 2:
                print(
                    "\nAtomic Valencies and Mayer Atomic"
                    " Bond Orders found, reading data")
            bo = np.zeros((molecule.num_atoms(), molecule.num_atoms()))
            columns = ""
            while True:
                read_buffer = f.__next__()
                # Check if the whole line is integers only (Header line)
                if is_int("".join(read_buffer.split())) is True:
                    # And use this information to label the columns
                    columns = read_buffer.split()
                # If we get to the Löwdin charges, we're done reading
                elif read_buffer.find("Lowdin Atomic Charges") != -1:
                    break
                else:
                    row = read_buffer.split()
                    j = 1
                    for i in columns:
                        j = j + 1
                        bo[int(row[0]) - 1][int(i) - 1] = float(row[j])
    f.close()
    # if verbosity >= 3:
    #     print("\nBond Orders:")
    #     np.set_printoptions(suppress=True)
    #     np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
    #     print(bo)
    build_bond_orders(molecule, bo, verbosity=verbosity)


def read_orca_bond_orders(filename, molecule, verbosity=0):
    bo = []
    f = open(filename, 'r')
    for line in f:
        if line.find("Mayer bond orders larger than 0.1") != -1:
            bo = np.zeros((molecule.num_atoms(), molecule.num_atoms()))
            if verbosity >= 2:
                print(
                    "\nMayer bond orders larger than 0.1 found, reading data")
            bo = np.zeros((molecule.num_atoms(), molecule.num_atoms()))
            while True:
                read_buffer = f.__next__()
                # Check if the whole line isn't empty (in that case we're done)
                if read_buffer and read_buffer.strip():
                    # Break the line into pieces
                    read_buffer = read_buffer[1:].strip()
                    read_buffer = read_buffer.split("B")
                    for i in read_buffer:
                        bondpair1 = int(i[1:4].strip())
                        bondpair2 = int(i[8:11].strip())
                        order = i[-9:].strip()
                        bo[bondpair1][bondpair2] = order
                        bo[bondpair2][bondpair1] = order
                else:
                    break
    f.close()
    # if verbosity >= 3:
    #     print("\nBond Orders:")
    #     np.set_printoptions(suppress=True)
    #     np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
    #     print(bo)
    build_bond_orders(molecule, bo, verbosity=verbosity)


def read_xyz_coord(filename, molecule, verbosity=0):
    geom = []
    f = open(filename, 'r')
    # Now examine every line until a line that doesn't conform to the
    #  template or EOF is found
    mark_for_delete = 0
    for line in f:
        # Check if line conforms to template
        read_buffer = line.split()
        if len(read_buffer) != 4:
            # If there are not *exactly* four entries on this line, we mark
            #  the current geometry for deletion (but delete only if we find
            #  more atoms).
            mark_for_delete = 1
        else:
            if mark_for_delete == 1:
                del geom[:]
                mark_for_delete = 0
            if is_atom_symbol(read_buffer[0]) and is_float(read_buffer[1])\
                    and is_float(read_buffer[2]) and is_float(read_buffer[3]):
                geom.append(read_buffer)
                if verbosity >= 3:
                    if len(geom) == 1:
                        print(
                            "New structure found, starting to read structure")
                        print_found_atom(
                            read_buffer[0], float(read_buffer[1]),
                            float(read_buffer[2]),
                            float(read_buffer[3]))
                    else:
                        print_found_atom(
                            read_buffer[0], float(read_buffer[1]),
                            float(read_buffer[2]),
                            float(read_buffer[3]))
            else:
                # If the entries aren't an atomic symbol and 3 coords, we
                # delete the current geometry and start fresh.
                mark_for_delete = 1

    if verbosity >= 2:
        print("\nReading of geometry finished.")
        print("\nAdding atoms to WellFARe molecule: ", molecule.name)
    for i in geom:
        molecule.add_atom(Atom(sym=i[0], x=float(i[1]), y=float(i[2]),
                               z=float(i[3])))
        if verbosity >= 3:
            print_add_atom(i[0], i[1], i[2], i[3])
    f.close()


def read_adf_coord(filename, molecule, verbosity=3):
    geom = []
    f = open(filename, 'r')
    for line in f:
        # if line.find(" Coordinates (Cartesian)") != -1:
        #     if verbosity >= 2:
        #         print("\nCartesian Coordinates found")
        #     del geom[:]
        #     for i in range(0, 5):
        #         read_buffer = f.__next__()
        #     while True:
        #         read_buffer = f.__next__()
        #         if read_buffer.find("-----------") == -1:
        #             read_buffer = read_buffer.split()
        #             geom.append([read_buffer[1],
        #                     float(read_buffer[5]), float(read_buffer[6]),
        #                     float(read_buffer[7])])
        #             if verbosity >= 3:
        #                 print_found_atom(geom[-1][0], geom[-1][1], geom[-1][2],
        #                                  geom[-1][3])
        #         else:
        #             break
        if line.find(" FRAGMENTS") != -1:
            if verbosity >= 2:
                print("\nCartesian Coordinates found")
            del geom[:]
            # Move 2 lines forward, the data begins there
            for i in range(0, 2):
                read_buffer = f.__next__()
            while True:
                read_buffer = f.__next__()
                if read_buffer.isspace() is True:
                    break
                else:
                    read_buffer = read_buffer.split()
                    geom.append([read_buffer[1],
                                 float(read_buffer[4]), float(read_buffer[5]),
                                 float(read_buffer[6])])
                    if verbosity >= 3:
                        print_found_atom(geom[-1][0], geom[-1][1], geom[-1][2],
                                         geom[-1][3])

    if verbosity >= 2:
        print("\nReading of geometry finished.")
        print("\nAdding atoms to WellFARe molecule: ", molecule.name)
    for read_buffer in geom:
        molecule.add_atom(
            Atom(sym=read_buffer[0], x=read_buffer[1],
                 y=read_buffer[2],
                 z=read_buffer[3]))
        if verbosity >= 3:
            print_add_atom(read_buffer[0],
                           read_buffer[1],
                           read_buffer[2],
                           read_buffer[3])
    f.close()


def read_turbo_coord(filename, molecule, verbosity=0):
    # Reading from Turbomole's aoforce file
    geom = []
    f = open(filename, 'r')
    for line in f:
        if line.find(
                "atomic coordinates            atom    charge  isotop") != -1:
            if verbosity >= 2:
                print("\nCartesian Coordinates found")
            del geom[:]
            while True:
                read_buffer = f.__next__()
                if read_buffer and read_buffer.strip():
                    geom.append(read_buffer)
                    if verbosity >= 3:
                        read_buffer = read_buffer.split()
                        print_found_atom(
                            number_to_symbol[int(float(read_buffer[4]))],
                            bohr_to_ang(float(read_buffer[0])),
                            bohr_to_ang(float(read_buffer[1])),
                            bohr_to_ang(float(read_buffer[2])))
                else:
                    break
    if verbosity >= 2:
        print("\nReading of geometry finished.")
        print("\nAdding atoms to WellFARe molecule: ", molecule.name)
    for i in geom:
        read_buffer = i.split()
        molecule.add_atom(Atom(charge=int(float(read_buffer[4])),
                               x=bohr_to_ang(float(read_buffer[0])),
                               y=bohr_to_ang(float(read_buffer[1])),
                               z=bohr_to_ang(float(read_buffer[2]))))
        if verbosity >= 3:
            print_add_atom(
                number_to_symbol[int(float(read_buffer[4]))],
                bohr_to_ang(float(read_buffer[0])),
                bohr_to_ang(float(read_buffer[1])),
                bohr_to_ang(float(read_buffer[2])))
    f.close()


def read_orca_coord(filename, molecule, verbosity=0):
    geom = []
    f = open(filename, 'r')
    for line in f:
        if line.find("CARTESIAN COORDINATES (ANGSTROEM)") != -1:
            if verbosity >= 2:
                print("\nCartesian Coordinates found")
            del geom[:]
            read_buffer = f.__next__()
            while True:
                read_buffer = f.__next__()
                if read_buffer and read_buffer.strip():
                    geom.append(read_buffer)
                    if verbosity >= 3:
                        read_buffer = read_buffer.split()
                        print_found_atom(read_buffer[0], float(read_buffer[1]),
                                         float(read_buffer[2]),
                                         float(read_buffer[3]))
                else:
                    break
    if verbosity >= 2:
        print("\nReading of geometry finished.")
        print("\nAdding atoms to WellFARe molecule: ", molecule.name)
    for i in geom:
        read_buffer = i.split()
        molecule.add_atom(
            Atom(sym=read_buffer[0], x=float(read_buffer[1]),
                 y=float(read_buffer[2]),
                 z=float(read_buffer[3])))
        if verbosity >= 3:
            print_add_atom(read_buffer[0],
                           read_buffer[1],
                           read_buffer[2],
                           read_buffer[3])
    f.close()


def read_gauss_coord(filename, molecule, verbosity=0):
    geom = []
    f = open(filename, 'r')
    for line in f:
        if line.find("Input orientation:") != -1:
            if verbosity >= 2:
                print("\nInput orientation found, reading coordinates")
            del geom[:]
            for i in range(0, 4):
                read_buffer = f.__next__()
            while True:
                read_buffer = f.__next__()
                if read_buffer.find("-----------") == -1:
                    geom.append(read_buffer)
                    if verbosity >= 3:
                        read_buffer = read_buffer.split()
                        print_found_atom(
                            number_to_symbol[int(read_buffer[1])],
                            float(read_buffer[3]), float(read_buffer[4]),
                            float(read_buffer[5]))
                else:
                    break
    f.close()
    if len(geom) == 0:
        # For some reason, we don't have the "input orientation" printed,
        #  let's try and read from the standard orientation instead
        f = open(filename, 'r')
        for line in f:
            if line.find("Standard orientation:") != -1:
                if verbosity >= 2:
                    print(
                        "\nStandard orientation found, reading coordinates")
                del geom[:]
                for i in range(0, 4):
                    read_buffer = f.__next__()
                while True:
                    read_buffer = f.__next__()
                    if read_buffer.find("-----------") == -1:
                        geom.append(read_buffer)
                        if verbosity >= 3:
                            read_buffer = read_buffer.split()
                            print_found_atom(
                                number_to_symbol[int(read_buffer[1])],
                                float(read_buffer[3]), float(read_buffer[4]),
                                float(read_buffer[5]))
                    else:
                        break
        f.close()
    if verbosity >= 2:
        print("\nReading of geometry finished.")
        print("\nAdding atoms to WellFARe molecule: ", molecule.name)
    for i in geom:
        read_buffer = i.split()
        molecule.add_atom(
            Atom(charge=int(read_buffer[1]), x=float(read_buffer[3]),
                 y=float(read_buffer[4]),
                 z=float(read_buffer[5])))
        if verbosity >= 3:
            print_add_atom(
                number_to_symbol[int(read_buffer[1])], read_buffer[3],
                read_buffer[4], read_buffer[5])
    return


def read_gauss_qm_energy(filename, molecule, verbosity=0):
    f = open(filename, 'r')
    qm_energies = []
    for line in f:
        if line.find("SCF Done:") != -1:
            readBuffer = line.split()
            readBuffer = float(readBuffer[4])
            qm_energies.append(readBuffer)
            if verbosity >= 3:
                print("\nEnergy from SCF cycle found")
                print("SCF:", str(readBuffer))
        elif line.find(" EUMP2 = ") != -1:
            readBuffer = line.split()
            readBuffer = float(readBuffer[5].replace('D', 'E'))
            qm_energies.append(readBuffer)
            if verbosity >= 3:
                print("\nEnergy from MP2 calculation found")
                print("MP2:", str(readBuffer))
        elif line.find(" CCSD(T)= ") != -1:
            readBuffer = line.split()
            readBuffer = float(readBuffer[1].replace('D', 'E'))
            qm_energies.append(readBuffer)
            if verbosity >= 3:
                print("\nEnergy from CCSD(T) calculation found")
                print("CCSD(T):", str(readBuffer))

    # Take the last QM energy from the list and assign that value as the
    # QM equilibrium energy of the molecule
    qm_energy = qm_energies[len(qm_energies) - 1]
    molecule.qm_energy = qm_energy
    if verbosity >= 2:
        print("\nReading of QM equilibrium energy complete")
        if verbosity >= 3:
            print("Ee_QM = " + str(qm_energy))
    f.close()


def read_orca_qm_energy(filename, molecule, verbosity=0):
    f = open(filename, 'r')
    qm_energies = []
    for line in f:
        if line.find("FINAL SINGLE POINT ENERGY") != -1:
            if verbosity >= 3:
                print("\nSingle point energy found")
                print(str(line))
            qm_energies.append(line)
    i = qm_energies[len(qm_energies) - 1]
    read_buffer = i.split()
    qm_energy_on_file = float(read_buffer[4])
    molecule.qm_energy = qm_energy_on_file
    if verbosity >= 2:
        print("\nReading of QM equilibrium energy complete")
        if verbosity >= 3:
            print("Ee_QM = " + str(qm_energy_on_file))
    f.close()


def read_turbomole_qm_energy(filename, molecule, verbosity=0):
    f = open(filename, 'r')
    qm_energies = []
    for line in f:
        if line.find("         *    SCF-energy") != -1:
            if verbosity >= 3:
                print("\nSCF energy found")
                print(str(line))
            qm_energies.append(line)
    i = qm_energies[len(qm_energies) - 1]
    read_buffer = i.split()
    qm_energy_on_file = float(read_buffer[3])
    molecule.qm_energy = qm_energy_on_file
    if verbosity >= 2:
        print("\nReading of QM equilibrium energy complete")
        if verbosity >= 3:
            print("Ee_QM = " + str(qm_energy_on_file))
    f.close()


def read_adf_qm_energy(filename, molecule, verbosity=0):
    f = open(filename, 'r')
    qm_energies = []
    for line in f:
        if line.find("  Total Bonding Energy:") != -1:
            if verbosity >= 3:
                print("\n Total Bonding Energy: found")
                print(str(line))
            qm_energies.append(line)
    i = qm_energies[len(qm_energies) - 1]
    read_buffer = i.split()
    qm_energy_on_file = float(read_buffer[3])
    molecule.qm_energy = qm_energy_on_file
    if verbosity >= 2:
        print("\nReading of QM equilibrium energy complete")
        if verbosity >= 3:
            print("Ee_QM = " + str(qm_energy_on_file))
    f.close()


def read_gauss_force_constants(filename, molecule, verbosity=0):
    H = []
    f = open(filename, 'r')
    for line in f:
        if line.find("Force constants in Cartesian coordinates") != -1:
            if verbosity >= 2:
                print(
                    "\nForce constants in Cartesian coordinates, reading data")
            H = np.zeros((3 * molecule.num_atoms(), 3 * molecule.num_atoms()))
            while True:
                readBuffer = f.__next__()
                # Check if the whole line is integers only (Header line)
                if is_int("".join(readBuffer.split())) == True:
                    # And use this information to label the columns
                    columns = readBuffer.split()
                # Once we find one of these statements, we're done reading
                elif readBuffer.find(
                        "FormGI is forming") != -1 or readBuffer.find(
                        "Cartesian forces in FCRed") != -1 or readBuffer.find(
                    "Final forces over variables") != -1:
                    break
                else:
                    row = readBuffer.split()
                    for i in range(0, len(row) - 1):
                        H[int(row[0]) - 1][int(columns[i]) - 1] = row[
                            i + 1].replace('D', 'E')
                        H[int(columns[i]) - 1][int(row[0]) - 1] = row[
                            i + 1].replace('D', 'E')
    # Store H as the QM calculated Hessian for this molecule
    molecule.set_hessian(H)
    if verbosity >= 3:
        print(
            "\nForce constants in Cartesian coordinates (Input orientation):")
        # np.set_printoptions(suppress=True)
        # np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
        print(H)
    f.close()


def read_orca_force_constants(filename, molecule, verbosity=0):
    H = np.zeros((3 * molecule.num_atoms(), 3 * molecule.num_atoms()))
    molecule.set_hessian(H)
    if verbosity >= 3:
        print(
            "\nForce constants in Cartesian coordinates:")
        # np.set_printoptions(suppress=True)
        # np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
        print(H)


def read_turbo_force_constants(filename, molecule, verbosity=0):
    H = []
    f = open(filename, 'r')
    for line in f:
        if line.find("translational & rotational space projected out") != -1:
            if verbosity >= 2:
                print(
                    "\nForce constants in Cartesian coordinates, reading data")
            H = np.zeros((3 * molecule.num_atoms(), 3 * molecule.num_atoms()))
            for i in range(0, 2):
                readBuffer = f.__next__()
            while True:
                readBuffer = f.__next__()
                # Check if the line is labeled: "ATOM" (Header line)
                if readBuffer.find("ATOM") != -1:
                    # And use this information to label the columns
                    columns = []
                    columns.append(int(readBuffer.split()[1]))
                    if len(readBuffer.split()) > 3:
                        columns.append(int(readBuffer.split()[3]))
                # Look for the lines below the header lines
                elif readBuffer.find(
                        "dx        dy        dz") != -1 or readBuffer.isspace() == True:
                    # And then we do nothing with this line
                    pass
                # Look for lines that begin with an atom number
                elif len(readBuffer.split()) > 0 and is_int(readBuffer.split()[0]) == True:
                    # And then we use that bit to set the row number
                    row = int(readBuffer.split()[0])
                    rowAdd = 0
                    readRawData = readBuffer.replace("-", " -")
                    readData = readRawData.split()[3:]
                    for j, i in enumerate(readData):
                        H[(row - 1) * 3][((columns[0] - 1) * 3) + j] = i
                        H[((columns[0] - 1) * 3) + j][(row - 1) * 3] = i
                # Once we find one of these statements, we're done reading
                elif readBuffer.find(
                        "*** projected hessian written onto") != -1:
                    break
                else:
                    rowAdd += 1
                    readRawData = readBuffer.replace("-", " -")
                    readData = readRawData.split()[1:]
                    for j, i in enumerate(readData):
                        H[((row - 1) * 3) + rowAdd][
                            ((columns[0] - 1) * 3) + j] = i
                        H[((columns[0] - 1) * 3) + j][
                            ((row - 1) * 3) + rowAdd] = i

    molecule.set_hessian(H)  # Store H as the QM calculated Hessian for this molecule
    if verbosity >= 3:
        print(
            "\nForce constants in Cartesian coordinates (translational & rotational space projected out):")
        # np.set_printoptions(suppress=True)
        # np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
        print(H)
    f.close()


def read_adf_force_constants(filename, molecule, verbosity=0):
    H = []
    f = open(filename, 'r')
    for line in f:
        if line.find("HESSIAN : AFTER SYM EXPAND & DIAGONAL CORRECTION") != -1:
            if verbosity >= 2:
                print(
                    "\nForce constants in Cartesian coordinates, reading data")
            # skip 3 lines ahead, that's where the data starts
            for i in range(0, 3):
                readBuffer = f.__next__()
            while True:
                readBuffer = f.__next__()
                if readBuffer.find("=====") != -1:
                    break
                else:
                    h_line=[float(i) for i in readBuffer.split()[3:]]
                    H.append(h_line)
    # Store H as the QM calculated Hessian for this molecule
    molecule.set_hessian(H)
    if verbosity >= 3:
        print(
            "\nForce constants in Cartesian coordinates (Input orientation):")
        # np.set_printoptions(suppress=True)
        # np.set_printoptions(formatter={'float': '{: 0.3f}'.format})
        print(H)
    f.close()


def read_gauss_frequencies(filename, molecule, verbosity=0):
    freq = []
    f = open(filename, 'r')
    for line in f:
        # The line below resets the frequency list in files that contain more than one frequency calculation
        if line.find(" Harmonic frequencies (cm**-1),") != -1:
            freq = []
        # The line below actually reads the frequencies
        if line.find(" Frequencies -- ") != -1:
            if verbosity >= 2:
                print("\nReading frequencies from file.")
            readBuffer = line.split()
            for i in readBuffer[2:]:
                freq.append(float(i))
    if verbosity >= 3:
        print("\nFrequencies:")
        print(freq)
    f.close()
    molecule.frequencies = freq


def read_orca_frequencies(filename, molecule, verbosity=0):
    freq = []
    f = open(filename, 'r')
    for line in f:
        if line.find("VIBRATIONAL FREQUENCIES") != -1:
            if verbosity >= 2:
                print("\nReading frequencies from file.")
            while True:
                readBuffer = f.__next__()
                # Check if the whole line is integers only (Header line)
                if readBuffer.find("cm**-1") != -1:
                    if math.fabs(float(readBuffer.split()[1])) > 1.0:
                        freq.append(float(readBuffer.split()[1]))
                # Once we find this statement, we're done reading
                elif readBuffer.find("NORMAL MODES") != -1:
                    break
    if verbosity >= 3:
        print("\nFrequencies:")
        print(freq)
    f.close()
    molecule.frequencies = freq


def read_turbo_frequencies(filename, molecule, verbosity=0):
    freq = []
    f = open(filename, 'r')
    for line in f:
        if line.find("VIBRATIONAL FREQUENCIES") != -1:
            if verbosity >= 2:
                print("\nReading frequencies from file.")
            while True:
                readBuffer = f.__next__()
                # Check if the whole line is integers only (Header line)
                if readBuffer.find("cm**-1") != -1:
                    if math.fabs(float(readBuffer.split()[1])) > 1.0:
                        freq.append(float(readBuffer.split()[1]))
                # Once we find this statement, we're done reading
                elif readBuffer.find("NORMAL MODES") != -1:
                    break
    if verbosity >= 3:
        print("\nFrequencies:")
        print(freq)
    f.close()
    molecule.frequencies = freq


def read_adf_frequencies(filename, molecule, verbosity=0):
    freq = []
    f = open(filename, 'r')
    for line in f:
        if line.find("List of All Frequencies:") != -1:
            if verbosity >= 2:
                print("\nReading frequencies from file.")
            # skip 8 lines ahead, that's where the data starts
            for i in range(0,8):
                readBuffer = f.__next__()
            while True:
                readBuffer = f.__next__()
                # Check if the whole line is integers only (Header line)
                if readBuffer.isspace():
                    break
                else:
                    freq.append(float(readBuffer.split()[0]))
    if verbosity >= 3:
        print("\nFrequencies:")
        print(freq)
    f.close()
    molecule.frequencies = freq


def read_gauss_multiplicity(filename, molecule, verbosity=0):
    QM_multiplicities = []
    f = open(filename, 'r')
    for line in f:
        if line.find(" Multiplicity = ") != -1:
            if verbosity >= 3:
                print("\nMultiplicity found")
                print(str(line))
            QM_multiplicities.append(line)
    # Take the last multiplicity from the file and assign that value as the multiplicity of the molecule
    i = QM_multiplicities[len(QM_multiplicities) - 1]
    readBuffer = i.split()
    QMmult = int(readBuffer[5])
    molecule.set_mult(QMmult)
    if verbosity >= 2:
        print("\nReading of multiplicities complete")
        if verbosity >= 3:
            print("Multiplicity = " + str(QMmult))
    f.close()


def read_orca_multiplicity(filename, molecule, verbosity=0):
    QM_multiplicities = []
    f = open(filename, 'r')
    for line in f:
        if line.find(" Multiplicity           Mult") != -1:
            if verbosity >= 3:
                print("\nMultiplicity found")
                print(str(line))
            QM_multiplicities.append(line)
    # Take the last multiplicity from the file and assign that value as the multiplicity of the molecule
    i = QM_multiplicities[len(QM_multiplicities) - 1]
    readBuffer = i.split()
    QMmult = int(readBuffer[3])
    molecule.set_mult(QMmult)
    if verbosity >= 2:
        print("\nReading of multiplicities complete")
        if verbosity >= 3:
            print("Multiplicity = " + str(QMmult))
    f.close()


def read_turbo_multiplicity(filename, molecule, verbosity=0):
    if verbosity >= 3:
        print(" Reading multiplicity from turbomole file not implemented: assuming singlet.")
        molecule.set_mult(1)
    if verbosity >= 3:
        print(" Multiplicity = " + str(1))


def read_adf_multiplicity(filename, molecule, verbosity=0):
    QM_multiplicities = []
    f = open(filename, 'r')
    for line in f:
        if line.find("  Spin polar:") != -1:
            if verbosity >= 3:
                print("\nSpin polarisation (multiplicity) found")
                print(str(line))
            QM_multiplicities.append(line)
    if verbosity >= 2:
        print("\nReading of multiplicities complete")
    # Take the last multiplicity from the file and assign that value as
    # the multiplicity of the molecule. If none was found: singlet
    if len(QM_multiplicities) > 0:
        i = QM_multiplicities[len(QM_multiplicities) - 1]
        readBuffer = i.split()
        # Need to add "1" here to get from spin polarisation to multiplicity
        QMmult = int(float(readBuffer[4])) + 1
    else:
        QMmult = 1
        if verbosity >= 3:
            print(" Couldn't read multiplicity from file, assuming singlet.")
    molecule.set_mult(QMmult)
    if verbosity >= 3:
        print(" Multiplicity = " + str(QMmult))
    f.close()


def read_gauss_rotational_symmetry_number(filename, molecule, verbosity=0):
    QM_sigmaRotnums = []
    f = open(filename, 'r')
    for line in f:
        if line.find(" Rotational symmetry number ") != -1:
            if verbosity >= 3:
                print("\n Rotational symmetry number")
                print(str(line))
            QM_sigmaRotnums.append(line)
    # Take the last multiplicity from the file and assign that value as the multiplicity of the molecule
    if len(QM_sigmaRotnums) != 0:
        i = QM_sigmaRotnums[len(QM_sigmaRotnums) - 1]
        readBuffer = i.split()
        QM_sigmaRotnum = int(float(readBuffer[3]))
    else:
        if verbosity >= 3:
            print(
                "\n No rotational symmetry number found (this is probably an atom), using default value of 1")
        QM_sigmaRotnum = 1
    molecule.sigmaRot = QM_sigmaRotnum
    if verbosity >= 2:
        print(" Rotational symmetry number = " + str(QM_sigmaRotnum))
    f.close()


def read_orca_rotational_symmetry_number(filename, molecule, verbosity=0):
    QM_sigmaRotnums = []
    f = open(filename, 'r')
    for line in f:
        if line.find("sn is the rotational symmetry number. We have assumed ") != -1:
            if verbosity >= 3:
                print("\n Rotational symmetry number found")
                print(str(line))
            QM_sigmaRotnums.append(line)
    # Take the last multiplicity from the file and assign that value as the multiplicity of the molecule
    if len(QM_sigmaRotnums) != 0:
        i = QM_sigmaRotnums[len(QM_sigmaRotnums) - 1]
        read_buffer = i.split()
        QM_sigmaRotnum = int(float(read_buffer[9]))
    else:
        if verbosity >= 3:
            print(
                "\n No rotational symmetry number found (this is probably an atom), using default value of 1")
        QM_sigmaRotnum = 1
    molecule.sigmaRot = QM_sigmaRotnum
    if verbosity >= 2:
        print(" Rotational symmetry number = " + str(QM_sigmaRotnum))
    f.close()


def read_turbo_rotational_symmetry_number(filename, molecule, verbosity=0):
    QM_sigmaRotnums = []
    #TODO: Find a way to determine from Turbomole output or write own module.


def read_adf_rotational_symmetry_number(filename, molecule, verbosity=0):
    QM_sigmaRotnums = []
    f = open(filename, 'r')
    for line in f:
        if line.find(" reported below were computed using sigma = ") != -1:
            if verbosity >= 3:
                print("\n Rotational symmetry number found")
                print(str(line))
            QM_sigmaRotnums.append(line)
    # Take the last multiplicity from the file and assign that value as the multiplicity of the molecule
    if len(QM_sigmaRotnums) != 0:
        i = QM_sigmaRotnums[len(QM_sigmaRotnums) - 1]
        read_buffer = i.split()
        QM_sigmaRotnum = int(float(read_buffer[7].replace(",","")))
    else:
        if verbosity >= 3:
            print(
                "\n No rotational symmetry number found (this is probably an atom), using default value of 1")
        QM_sigmaRotnum = 1
    molecule.sigmaRot = QM_sigmaRotnum
    if verbosity >= 2:
        print(" Rotational symmetry number = " + str(QM_sigmaRotnum))
    f.close()


def extract_molecular_data(filename, molecule, verbosity=0,
                           read_coordinates=True, read_bond_orders=True,
                           read_qm_energy=False,
                           build_angles=False, build_dihedrals=False,
                           read_force_constants=False,
                           read_multiplicity=False,
                           read_rotational_symmetry_number=False,
                           cpu_number=1):
    if verbosity >= 1:
        print("\nSetting up WellFARe molecule {} from file {}.".format(
            molecule.name, filename))

    # Try filename for readability first
    try:
        f = open(filename, 'r')
        f.close()
    except OSError:
        msg_program_error("Cannot open " + filename + " for reading.")

    # Next, establish which kind of file we're dealing with
    # Determine which QM program we're dealing with
    f = open(filename, 'r')
    program = "N/A"
    for line in f:
        if line.find("Entering Gaussian System, Link 0") != -1:
            if verbosity >= 2:
                print("Reading Gaussian output file: ", filename)
            program = "gauss"
            break
        elif line.find("* O   R   C   A *") != -1:
            if verbosity >= 2:
                print("Reading Orca output file: ", filename)
            program = "orca"
            break
        elif line.find("T U R B O M O L E") != -1:
            if verbosity >= 2:
                print("Reading Turbomole output file: ", filename)
            program = "turbomole"
            break
        elif line.find("*   Amsterdam Density Functional  (ADF)") != -1:
            if verbosity >= 2:
                print("Reading ADF output file: ", filename)
            program = "adf"
            break
    f.close()
    if program == "N/A":
        if verbosity >= 2:
            print("Reading xyz file: ", filename)
        if verbosity >= 3:
            print("(Note that this is also a fallback)")
        program = "xyz"

    # Check if we need to read coordinates
    if read_coordinates is True:
        if program == "gauss":
            read_gauss_coord(filename, molecule, verbosity=verbosity)
        elif program == "orca":
            read_orca_coord(filename, molecule, verbosity=verbosity)
        elif program == "turbomole":
            read_turbo_coord(filename, molecule, verbosity=verbosity)
        elif program == "adf":
            read_adf_coord(filename, molecule, verbosity=verbosity)
        else:
            read_xyz_coord(filename, molecule, verbosity=verbosity)

    # Check if we need to read bond orders
    if read_bond_orders is True:
        if program == "gauss":
            read_gauss_bond_orders(filename, molecule, verbosity=verbosity)
        elif program == "orca":
            read_orca_bond_orders(filename, molecule, verbosity=verbosity)
        elif program == "turbomole":
            build_bond_orders(molecule, [], verbosity=verbosity)
        else:
            build_bond_orders(molecule, [], verbosity=verbosity,
                                       cpu_number=cpu_number)

    # Check if we need to read the quantum mechanically calculated energy
    if read_qm_energy is True:
        if program == "gauss":
            read_gauss_qm_energy(filename, molecule, verbosity=verbosity)
        elif program == "orca":
            read_orca_qm_energy(filename, molecule, verbosity=verbosity)
        elif program == "turbomole":
            read_turbomole_qm_energy(filename, molecule, verbosity=verbosity)
        elif program == "adf":
            read_adf_qm_energy(filename, molecule, verbosity=verbosity)
        else:
            # There's no QM energy in xyz files...
            pass

    # Check if we need to build bond angles
    if build_angles is True:
        build_molecular_angles(molecule, verbosity=verbosity)

    # Check if we need to build dihedrals
    if build_dihedrals is True:
        build_molecular_dihedrals(molecule, verbosity=verbosity)

    # Check if we need to read (harmonic) force constants. If, after that, the
    # matrix of force constants is still empty, we'll try to read frequencies.
    if read_force_constants is True:
        if program == "gauss":
            read_gauss_force_constants(filename, molecule, verbosity=verbosity)
            if molecule.H_QM == []:
                read_gauss_frequencies(filename, molecule, verbosity=verbosity)
        elif program == "orca":
            read_orca_force_constants(filename, molecule, verbosity=verbosity)
            if molecule.H_QM == []:
                read_orca_frequencies(filename, molecule, verbosity=verbosity)
        elif program == "turbomole":
            read_turbo_force_constants(filename, molecule, verbosity=verbosity)
            if molecule.H_QM == []:
                read_turbo_frequencies(filename, molecule, verbosity=verbosity)
        elif program == "adf":
            read_adf_force_constants(filename, molecule, verbosity=verbosity)
            if molecule.H_QM == []:
                read_adf_frequencies(filename, molecule, verbosity=verbosity)
        else:
            # There are no force constants in xyz files...
            pass

    # Check if we need to read multiplicities
    if read_multiplicity is True:
        if program == "gauss":
            read_gauss_multiplicity(filename, molecule,
                                       verbosity=verbosity)
        elif program == "orca":
            read_orca_multiplicity(filename, molecule,
                                      verbosity=verbosity)
        elif program == "turbomole":
            read_turbo_multiplicity(filename, molecule,
                                       verbosity=verbosity)
        elif program == "adf":
            read_adf_multiplicity(filename, molecule,
                                     verbosity=verbosity)
        else:
            # There is no multiplicity in an xyz file...
            pass

    # Check if we need to read rotational symmetry numbers
    if read_rotational_symmetry_number is True:
        if program == "gauss":
            read_gauss_rotational_symmetry_number(filename, molecule,
                                    verbosity=verbosity)
        elif program == "orca":
            read_orca_rotational_symmetry_number(filename, molecule,
                                   verbosity=verbosity)
        elif program == "turbomole":
            read_turbo_rotational_symmetry_number(filename, molecule,
                                    verbosity=verbosity)
        elif program == "adf":
            read_adf_rotational_symmetry_number(filename, molecule,
                                  verbosity=verbosity)
        else:
            # There are no rotational symmetry numbers in an xyz file...
            pass


def main():
    prg_start_time = time.time()
    # Create an example molecule and add some atoms
    example = Molecule("Example Molecule")
    # extract_molecular_data("../examples/g09-h3b-nh3.log", example, verbosity=3)
    extract_molecular_data("../examples/pdb200l.xyz", example, verbosity=3,
                           read_coordinates=True, read_bond_orders=True,
                           build_angles=True, build_dihedrals=True, cpu_number=4)
    # print(example.print_mol(output="gauss"))
    msg_program_footer(prg_start_time)


if __name__ == '__main__':
    main()



###############################################################################
# Methods for unit conversions are defined below
###############################################################################

from wellfare.constants import symbol_to_number


# Conversion of length in Angstroms to
# atomic units (Bohrs)
def ang_to_bohr(ang):
    return ang * 1.889725989


def bohr_to_ang(bohr):
    return bohr / 1.889725989


# Test if the argument is (can be converted to)
# an integer number
def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


# Test if the argument is (can be converted to)
# a floating point number
def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


# Test if the argument is (can be converted to)
# an atomic symbol
def is_atom_symbol(s):
    if s in symbol_to_number:
        return True
    else:
        return False



import math

import numpy as np

# from typing import Optional
#
# from molecule import Molecule
from wellfare.conversions import *
from wellfare.constants import *
from wellfare.messages import *


###############################################################################
#                                                                             #
# This is the part of the program where the vibrational analysis is done      #
#                                                                             #
###############################################################################

def thermochemical_analysis(molecule, temp=298.15, press=101325.0,
                            scalefreq=1.0, internal="truhlar", cutoff=100.0,
                            verbosity=0):
    # Constants
    # k_boltz() = 1.38064852E-23  # Boltzmann constant in JK⁻¹
    # h_planck() = 6.626070040E-34  # Planck constant in Js
    # r_gas() = 8.3144598  # Gas constant in J K⁻¹ mol⁻¹
    # c_light() = 2.9979258E10  # Speed of light in cm s⁻¹

    # Print T and p conditions
    if verbosity >= 2:
        print("\n Thermal analysis for WellFARe molecule: ", molecule.name)
        print(" Temperature: {:> 12.2f} K".format(temp))
        print(" Pressure   : {:> 12.2f} Pa".format(press))

    # Electronic contribution to the entropy
    Qelec = molecule.mult
    molecule.elecS = r_gas() * math.log(Qelec)
    if verbosity >= 2:
        print(
            "\n Electronic partition function:                  {:> 12.5E}".format(
                Qelec))
        print(
            " Electronic contribution to the entropy:         {:> 9.2f} J mol⁻¹ K⁻¹".format(
                molecule.elecS))

    # Translational Partition Function
    Qtrans = (((2 * math.pi * (
            molecule.mass() / 6.0221409E26) * k_boltz() * temp) / (
                       h_planck() ** 2)) ** (3 / 2)) * (
                     (k_boltz() * temp) / press)
    molecule.transS = r_gas() * (math.log(Qtrans) + 1 + (3 / 2))
    Etrans = (3 / 2) * r_gas() * temp
    molecule.thermTrans = Etrans / 2625500.2
    if verbosity >= 2:
        print(
            "\n Translational partition function:               {:> 12.5E}".format(
                Qtrans))
        print(
            " Translational contribution to the total energy: {:> 12.2f} J mol⁻¹".format(
                Etrans))
        print(
            " Translational contribution to the entropy:      {:> 12.2f} J mol⁻¹ K⁻¹".format(
                molecule.transS))

    # Rotational Partition Function
    # We need to distinguish three cases: (1) single atom, (2) linear molecule, (3) non-linear molecule
    # For cases (2) and (3), we also need to determine the correct symmetry number sigma_r

    if molecule.num_atoms() == 1:
        if verbosity >= 3:
            print(" No rotational contribution for single atoms.")
        # Case (1): single atom
        q_rot = 1.0
        molecule.rotS = 0.0
        e_rot = 0.0
    else:
        # Now we need to calculate the moments of inertia to determine if the molecule is linear or not
        # (and to guesstimate the symmetry number)
        # The molecular center of mass
        xValue = 0.0
        yValue = 0.0
        zValue = 0.0
        for i in molecule.atoms:
            xValue = xValue + (i.mass * i.coord[0])
            yValue = yValue + (i.mass * i.coord[1])
            zValue = zValue + (i.mass * i.coord[2])
        xValue = xValue / (molecule.mass())
        yValue = yValue / (molecule.mass())
        zValue = zValue / (molecule.mass())

        # Translate whole molecule into the center of mass reference frame
        for i in molecule.atoms:
            i.coord[0] = i.coord[0] - xValue
            i.coord[1] = i.coord[1] - yValue
            i.coord[2] = i.coord[2] - zValue

        # Build inertia tensor
        inertia_tensor = []
        Ixx = 0.0
        Ixy = 0.0
        Ixz = 0.0
        Iyx = 0.0
        Iyy = 0.0
        Iyz = 0.0
        Izx = 0.0
        Izy = 0.0
        Izz = 0.0
        for i in molecule.atoms:
            Ixx = Ixx + (
                    i.mass * (
                    (ang_to_bohr(i.coord[1]) * ang_to_bohr(i.coord[1])) + (
                    ang_to_bohr(i.coord[2]) * ang_to_bohr(i.coord[2]))))
            Ixy = Ixy - i.mass * ang_to_bohr(i.coord[0]) * ang_to_bohr(
                i.coord[1])
            Ixz = Ixz - i.mass * ang_to_bohr(i.coord[0]) * ang_to_bohr(
                i.coord[2])
            Iyx = Iyx - i.mass * ang_to_bohr(i.coord[1]) * ang_to_bohr(
                i.coord[0])
            Iyy = Iyy + (
                    i.mass * (
                    (ang_to_bohr(i.coord[0]) * ang_to_bohr(i.coord[0])) + (
                    ang_to_bohr(i.coord[2]) * ang_to_bohr(i.coord[2]))))
            Iyz = Iyz - i.mass * ang_to_bohr(i.coord[1]) * ang_to_bohr(
                i.coord[2])
            Izx = Izx - i.mass * ang_to_bohr(i.coord[2]) * ang_to_bohr(
                i.coord[0])
            Izy = Izy - i.mass * ang_to_bohr(i.coord[2]) * ang_to_bohr(
                i.coord[1])
            Izz = Izz + (
                    i.mass * (
                    (ang_to_bohr(i.coord[0]) * ang_to_bohr(i.coord[0])) + (
                    ang_to_bohr(i.coord[1]) * ang_to_bohr(i.coord[1]))))
        inertia_tensor.append([Ixx, Ixy, Ixz])
        inertia_tensor.append([Iyx, Iyy, Iyz])
        inertia_tensor.append([Izx, Izy, Izz])
        inertia_tensor = np.matrix(inertia_tensor)

        # Diagonalise inertia tensor
        inertia_moments, inertial_axes = np.linalg.eig(inertia_tensor)

        # Orthogonalise eigenvectors (only sometimes necessary)...
        inertial_axes, r = np.linalg.qr(inertial_axes)

        # Sort moments from highest to lowest
        idx = inertia_moments.argsort()[::-1]
        inertia_moments = inertia_moments[idx]
        inertial_axes = inertial_axes[:, idx]
        if verbosity >= 3:
            print("\n Moments of inertia in atomic units:")
            print(inertia_moments)

        if inertia_moments[0] - inertia_moments[1] < 0.0005:
            # Case (2): linear molecule
            if verbosity >= 3:
                print("\n Rotational contribution for a linear molecule.")
            phi = (h_planck() ** 2) / (8 * (math.pi ** 2) * (
                    inertia_moments[0] * 4.6468689E-48) * k_boltz())
            q_rot = (1 / molecule.sigmaRot) * (temp / phi)
            molecule.rotS = r_gas() * (math.log(q_rot) + 1)
            e_rot = r_gas() * temp
        else:
            # Case (3): nonlinear molecule
            if verbosity >= 3:
                print("\n Rotational contribution for a nonlinear molecule.")
            phi_x = (h_planck() ** 2) / (8 * (math.pi ** 2) * (
                    inertia_moments[0] * 4.6468689E-48) * k_boltz())
            phi_y = (h_planck() ** 2) / (8 * (math.pi ** 2) * (
                    inertia_moments[1] * 4.6468689E-48) * k_boltz())
            phi_z = (h_planck() ** 2) / (8 * (math.pi ** 2) * (
                    inertia_moments[2] * 4.6468689E-48) * k_boltz())
            q_rot = (math.sqrt(math.pi) / molecule.sigmaRot) * (
                    math.sqrt(temp ** 3) / math.sqrt(phi_x * phi_y * phi_z))
            molecule.rotS = r_gas() * (math.log(q_rot) + (3 / 2))
            e_rot = (3 / 2) * r_gas() * temp

    if verbosity >= 2:
        print(
            "\n Rotational partition function:                  {:> 12.5E}".format(
                q_rot))
        print(
            " Rotational contribution to the total energy:    {:> 12.2f} J mol⁻¹".format(
                e_rot))
        print(
            " Rotational contribution to the entropy:         {:> 12.2f} J mol⁻¹ K⁻¹\n".format(
                molecule.rotS))

    if molecule.num_atoms() != 1 and (
            molecule.H_QM != [] or molecule.frequencies != []):
        # If we didn't read frequencies from file, calculate them from force constants
        listOfFreqs = []
        if molecule.frequencies == []:
            if verbosity >= 2:
                print("\n Vibrational analysis:")
            # First, we need to form the mass-weighted force-constant matrix
            molecule.H_mw = np.zeros((len(molecule.H_QM), len(molecule.H_QM)))
            for i in range(0, molecule.num_atoms()):
                for j in range(0, molecule.num_atoms()):
                    for h in range(0, 3):
                        for l in range(0, 3):
                            molecule.H_mw[3 * i + h][3 * j + l] = \
                                molecule.H_QM[3 * i + h][
                                    3 * j + l] / math.sqrt(
                                    symbol_to_au_mass[
                                        molecule.atm_symbol(i)] *
                                    symbol_to_au_mass[
                                        molecule.atm_symbol(j)])
            frequencies, normalModes = np.linalg.eig(molecule.H_mw)

            # We need to distinguish two cases: (1) linear molecule with 3N-5 vibrations and (2) non-linear
            # molecules with 3N-6 vibrations.
            if inertia_moments[0] - inertia_moments[1] < 0.0005 and inertia_moments[2] < 0.0005:
                # Case (1): linear molecule
                if verbosity >= 3:
                    print(
                        "\n Vibrational contribution for a linear molecule (3N-5)")
                if verbosity >= 3:
                    print(
                        "\n Frequencies that correspond to rotation and translation (lowest 5):")
                    for i in np.sort(frequencies)[:5]:
                        conversion = 219474.63  # atomic units to cm⁻¹ in a harmonic oscillator
                        if i < 0.0:
                            sign = -1
                        else:
                            sign = 1
                        i = sign * math.sqrt(abs(i)) * conversion
                        print("{:> 9.2f} cm⁻¹".format(i))

                for i in np.sort(frequencies)[5:]:
                    conversion = 219474.63  # atomic units to cm⁻¹ in a harmonic oscillator
                    if i < 0.0:
                        sign = -1
                    else:
                        sign = 1
                    i = sign * math.sqrt(abs(i)) * conversion

                    # We don't add imaginary modes to the thermochemical analysis
                    if sign == 1:
                        listOfFreqs.append(i)
                # Then save the list to the molecule
                molecule.frequencies = listOfFreqs

            else:
                # Case (2): nonlinear molecule
                if verbosity >= 3:
                    print(
                        "\n Vibrational contribution for a nonlinear molecule (3N-6)")
                if verbosity >= 3:
                    print(
                        "\n Frequencies that correspond to rotation and translation (lowest 6):")
                    for i in np.sort(frequencies)[:6]:
                        conversion = 219474.63  # atomic units to cm⁻¹ in a harmonic oscillator
                        if i < 0.0:
                            sign = -1
                        else:
                            sign = 1
                        i = sign * math.sqrt(abs(i)) * conversion
                        print("{:> 9.2f} cm⁻¹".format(i))
                for i in np.sort(frequencies)[6:]:
                    conversion = 219474.63  # atomic units to cm⁻¹ in a harmonic oscillator
                    if i < 0.0:
                        sign = -1
                    else:
                        sign = 1
                    i = sign * math.sqrt(abs(i)) * conversion

                    # We don't add imaginary modes to the thermochemical analysis
                    if sign == 1:
                        listOfFreqs.append(i)
                # Then save the list to the molecule
                molecule.frequencies = listOfFreqs
            if verbosity >= 2:
                print(
                    "\n Frequencies (before contamination with rotation and translation has been removed):")
                if verbosity >= 3:
                    print(" (imaginary modes will be ignored)")
                for i in listOfFreqs:
                    print("{:> 9.2f} cm⁻¹".format(i))
        else:
            # If we've read the frequencies from file, just use those
            for i in molecule.frequencies:
                if i < 0.0:
                    # We don't add imaginary modes to the thermochemistry
                    pass
                else:
                    listOfFreqs.append(i)

        # Check if we have been asked to scale all frequencies
        if scalefreq != 1.0:
            if verbosity >= 1:
                print("\n Scaling all frequencies by a factor of {}".format(
                    scalefreq))
            for i in range(0, len(listOfFreqs)):
                if verbosity >= 1:
                    print(
                        " Vibration {} Before: {:> 9.2f} cm⁻¹, After: {:> 9.2f} cm⁻¹".format(
                            i + 1, listOfFreqs[i],
                            listOfFreqs[
                                i] * scalefreq))
                listOfFreqs[i] = listOfFreqs[i] * scalefreq

        # Following Truhlar's advice, very low frequencies (below
        # 100 cm⁻¹) should be fixed to exactly 100 cm⁻¹. We check
        # if this option is selected and then proceed:
        if internal == "truhlar":
            if verbosity >= 1:
                print("\n Fixing all low frequencies to {:> 6.1f}"
                      " cm⁻¹".format(cutoff))
            for i in range(0, len(listOfFreqs)):
                if 0.0 < listOfFreqs[i] < cutoff:
                    if verbosity >= 1:
                        print(
                            " Vibration {} Before: {:> 6.1f} cm⁻¹,"
                            " After: {:> 6.1f} cm⁻¹".format(i + 1, listOfFreqs[i],cutoff))
                    listOfFreqs[i] = cutoff
        # Alternatively, all low-lying modes can be simply ignored completely
        # Here, we make assumptions about which modes may have an outsize
        # contribution and completely delete the corresponding frequencies.
        elif internal == "ignorelow":
            if verbosity >= 1:
                print("\n Discarding all low frequencies below {:> 6.1f}"
                      " cm⁻¹".format(cutoff))
            # We're going through this list in reverse order, so we can delete
            # matching items straight away without messing up the list index
            for i in range(len(listOfFreqs)-1,-1,-1):
                if 0.0 < listOfFreqs[i] < cutoff:
                    if verbosity >= 1:
                        print(
                            " Vibration {} ({:> 6.1f} cm⁻¹),"
                            " discarded".format(i + 1, listOfFreqs[i]))
                        del listOfFreqs[i]

        if verbosity >= 2:
            print(
                "\n Frequencies (after scaling and internal rotor treatment):")
            if verbosity >= 3:
                print(" (imaginary modes will be ignored)")
            for i in listOfFreqs:
                print("{:> 9.2f} cm⁻¹".format(i))

        # Then, create a list that contains all vibrational temperatures (makes the summation of the
        # partition function simpler/more convenient)
        listOfVibTemps = []
        if verbosity >= 2:
            print("\n Vibrational temperatures:")
        ZPVE = 0.0
        for i in listOfFreqs:
            nu = i * c_light()
            phi = h_planck() * nu / k_boltz()
            # Still need to check for imaginary modes, since we might have just read the frequencies
            # instead of calculating them ourselves.
            if i > 0.0:
                molecule.ZPVEList.append(
                    (h_planck() * nu * 6.0221409E23) / 2.0 / 2625500.2)
                ZPVE += h_planck() * nu * 6.0221409E23
                listOfVibTemps.append(phi)
            if verbosity >= 2:
                print("{:> 9.2f} K".format(phi))
        molecule.ZPVE = (
                                ZPVE / 2.0) / 2625500.2  # Converted from Joules to Hartree straight away
        # Now, assemble the partition function, vibrational entropy and the vibrational contribution to
        # the total thermal energy along with the zero point vibrational energy
        Qvib = 1.0
        Svib = 0.0
        Evib = 0.0
        if verbosity >= 2 and internal == "grimme":
            print("\n Interpolation between entropic contributions from"
                  " harmonic oscillators and free rotors for low frequency"
                  " modes:")
        for j, i in enumerate(listOfVibTemps):
            Qvib = Qvib * (math.exp(-1.0 * i / (2.0 * temp))) / (
                    1 - math.exp(-1.0 * i / (temp)))
            if internal == "grimme":
                # Frequency of the vibration = freq of free rotor
                omega = listOfFreqs[j] * c_light()
                # Moment of the free rotor
                mu = h_planck()/(8*math.pi*math.pi*omega)
                # Reduced value for the actual moment to make sure its value
                # is within reasonable bounds
                B_av = (inertia_moments[0] + inertia_moments[1] +
                        inertia_moments[2]) / 3
                mu_prime = (mu * B_av)/(mu + B_av)
                # Vibrational entropy as usual
                vibrational_entropy = ((i / temp) / (
                        math.exp(i / (temp)) - 1.0)) - math.log(
                    1.0 - math.exp(-1.0 * i / (temp)))
                vibrational_entropy = vibrational_entropy * r_gas()
                # Rotational entropy of the free rotor
                rotational_entropy = (8 * (
                            math.pi ** 3) * mu_prime * k_boltz() * temp) / (
                                                 h_planck() * h_planck())
                rotational_entropy = math.log(math.sqrt(rotational_entropy))
                rotational_entropy = (0.5 + rotational_entropy) * r_gas()
                # Weighting function to balance between vib and rot entropy
                w = 1/(1+(((cutoff* c_light())/omega)**4))
                if verbosity >= 3:
                    print(" Interpolation for vibration {}: {}% Svib, {}% Srot.".format(j,int(w*100),int((1-w)*100)))
                SvibContribution = (w * vibrational_entropy) + (
                            (1 - w) * rotational_entropy)
            else:
                SvibContribution = ((i / temp) / (
                        math.exp(i / (temp)) - 1.0)) - math.log(
                    1.0 - math.exp(-1.0 * i / (temp)))
                SvibContribution = SvibContribution * r_gas()
            Svib += SvibContribution
            molecule.VibSList.append(SvibContribution * r_gas())
            EvibContribution = i * (0.5 + (1 / (math.exp(i / temp) - 1.0)))
            Evib += EvibContribution * r_gas()
            molecule.thermVibList.append(
                math.fabs((EvibContribution * r_gas()) - (
                        molecule.ZPVEList[j] * 2625500.2)) / 2625500.2)

    else:
        # If there are no vibrations, i.e. for a single atom, the results are straightforward:
        if verbosity >= 2:
            print(
                "\n No vibrational analysis. This is either"
                " a single atom, or we didn't find any data.")
        Qvib = 1.0
        Svib = 0.0
        molecule.vibS = 0.0
        Evib = 0.0

    molecule.vibS = Svib
    molecule.thermRot = (e_rot / 2625500.2)
    molecule.thermVib = (Evib / 2625500.2) - molecule.ZPVE
    Etherm = molecule.ZPVE + molecule.thermVib + molecule.thermRot + molecule.thermTrans
    Etot = molecule.qm_energy + molecule.ZPVE + molecule.thermVib + molecule.thermRot + molecule.thermTrans
    molecule.kT = (
            k_boltz() / 4.3597482E-18 * temp)  # kT contributon to the enthalpy
    Htot = Etot + molecule.kT  # Enthalpy: H = Etot + kB T in Hartree
    Stot = (
            molecule.elecS + molecule.transS + molecule.rotS + molecule.vibS)  # Total entropy
    molecule.negTS = (
            -1.0 * temp * Stot / 2625500.2)  # Total entropic contribution TS
    Gtot = Htot + molecule.negTS
    if molecule.frequencies == [] and molecule.H_QM == []:
        msg_program_warning(
            "\nCareful! We don't have any vibrational data for this"
            " compound!\nDon't use any of the derived thermodynamic"
            " values ΔE, ΔU, ΔH, ΔS or ΔG!\n")
    if verbosity >= 2:
        print(
            "\n Vibrational partition function:"
            "                  {:> 12.5E}\n".format(
                Qvib))
        print(
            " Vibrational contribution to the entropy:"
            "         {:> 12.2f} J mol⁻¹ K⁻¹\n".format(
                molecule.vibS))
        print("  Contributions to the ZPVE:")
        for i, j in enumerate(molecule.ZPVEList):
            print(
                "  Vibration {:4}:"
                "                        {:> 12.2f} J mol⁻¹"
                    .format(i + 1, j * 2625500.2))
        print(
            " Total zero-point vibrational energy (ZPVE):"
            "            {:> 12.2f} J mol⁻¹\n"
                .format(molecule.ZPVE * 2625500.2))
        print("  Contributions to the vibrations:")
        for i, j in enumerate(molecule.thermVibList):
            print(
                "  Vibration {:4}:"
                "                        {:> 12.2f} J mol⁻¹"
                    .format(i + 1, j * 2625500.2))
        print(
            " Total vibrational contribution to the energy:"
            "    {:> 12.2f} J mol⁻¹"
                .format(molecule.thermVib * 2625500.2))

    if verbosity >= 1:
        print("\n Summary of the thermochemistry for: ", molecule.name)
        if verbosity >= 2:
            print(
                " The temperature is {:> 6.2f} K and"
                " the pressure is {:> 10.2f} Pa for this analysis"
                    .format(temp, press))
        print(
            " Electronic Energy (Ee):"
            "                                 {:> 12.6f} h"
                .format(molecule.qm_energy))
        print(
            " Electronic Energy (Ee + ZPVE):"
            "                          {:> 12.6f} h"
                .format(molecule.qm_energy + molecule.ZPVE))
        if verbosity >= 2:
            print(
                "   Translational contribution:"
                "                           {:> 12.6f} h"
                    .format(molecule.thermTrans))
            print(
                "   Rotational contribution:"
                "                              {:> 12.6f} h"
                    .format(molecule.thermRot))
            print(
                "   Zero point vibrational contribution:"
                "                  {:> 12.6f} h"
                    .format(molecule.ZPVE))
            print(
                "   Finite temperature vibrational contribution:"
                "          {:> 12.6f} h"
                    .format(molecule.thermVib))
            print(
                "  Total thermal contributions to the energy (Etherm):"
                "    {:> 12.6f} h"
                    .format(Etherm))
        print(
            " Internal energy (U = Ee + Etherm):"
            "                      {:> 12.6f} h"
                .format(Etot))
        if verbosity >= 2:
            print(
                "  kT (=pV) contribution to the enthalpy:"
                "                 {:> 12.6f} h"
                    .format(molecule.kT))
        print(
            " Enthalpy (H = U + kT):"
            "                                  {:> 12.6f} h"
                .format(Htot))
        if verbosity >= 3:
            print(
                "   Electronic entropy:"
                "                               {:> 12.2f} J mol⁻¹ K⁻¹"
                    .format(molecule.elecS))
            print(
                "   Translational entropy:"
                "                            {:> 12.2f} J mol⁻¹ K⁻¹"
                    .format(molecule.transS))
            print(
                "   Rotational entropy:"
                "                               {:> 12.2f} J mol⁻¹ K⁻¹"
                    .format(molecule.rotS))
            print(
                "   Vibrational entropy:"
                "                              {:> 12.2f} J mol⁻¹ K⁻¹"
                    .format(molecule.vibS))
        if verbosity >= 2:
            print(
                "  Total entropy:"
                "                                     {:> 12.2f} J mol⁻¹ K⁻¹"
                    .format(Stot))
            print(
                " Total entropic contribution to the Gibbs energy (-TS)"
                "   {:> 12.6f} h\n"
                    .format(molecule.negTS))
        print(
            " Total Gibbs energy (G = H - TS)"
            "                         {:> 12.6f} h\n"
                .format(Gtot))



################################################################################
# Atom class and class methods to be defined below
################################################################################

from typing import Optional

from wellfare.constants import *


class Atom:
    """
    An atom defined by its nucleus (i.e. charge or symbol and (optional) mass)
    and its cartesian coordinates in Ångströms.
    """

    def __init__(self, sym: Optional[str] = None,
                 charge: Optional[int] = None, x: Optional[float] = 0.0,
                 y: Optional[float] = 0.0, z: Optional[float] = 0.0,
                 mass: Optional[float] = None) -> None:
        """
        This method creates an atom somewhere in cartesian space. Either charge
        (i.e. atomic number, number of protons, etc.) or the atomic symbol can
        be specified. Optionally, the atomic mass can be specified (in order to
        create a specific isotope/nuclide). Cartesian coordinates can be given
        to specify the location in space. By default, a hydrogen atom at the
        origin of the coordinate system is created.

        If both symbol and atomic number are given, only the symbol is used in
        the creation, the atomic number (consistent with the symbol or not) is
        discarded.

        Specification of an unknown atomic number will revert back to hydrogen.

        :param charge: The nuclear charge (i.e. the atomic number) that defines
                       the element.
        :param x: The cartesian x-coordinate of the atom in Ångströms.
        :param y: The cartesian y-coordinate of the atom in Ångströms.
        :param z: The cartesian z-coordinate of the atom in Ångströms.
        :param mass: Optionally, the mass of the atom in amu. If unspecified,
                     the standard atomic weight (as defined in constants.py
                     will be used)
        """

        if sym is None and charge is not None:
            if charge in number_to_symbol:
                self.charge = charge
            else:
                self.charge = 1
        elif sym is not None and charge is None:
            if sym in symbol_to_number:
                self.charge = symbol_to_number[sym]
            else:
                self.charge = 1
        elif sym is None and charge is None:
            self.charge = 1
        else:
            # If both are specified, the symbol takes precedence
            self.charge = symbol_to_number[sym]

        if mass is None:
            self.mass = symbol_to_mass[self.symbol()]
        else:
            self.mass = mass

        self.coord = [x, y, z]

    def __str__(self) -> str:
        """

        :return: Returns a string representation of this Atom in this
                 format: (MASS)SYMBOL, (X, Y, Z)
        """

        return '({: 8.3f}){}: ({: 13.8f}, {: 13.8f}, {: 13.8f}))'.format(self.mass,
                                                             self.symbol(),
                                                             self.coord[0],
                                                             self.coord[1],
                                                             self.coord[2])

    def __repr__(self) -> str:
        """

        :return: Returns a string representation of this Atom in this
                 format: (MASS)SYMBOL, (X, Y, Z)
        """

        return '({: 8.3f}){}: ({: 13.8f}, {: 13.8f}, {: 13.8f}))'.format(self.mass,
                                                             self.symbol(),
                                                             self.coord[0],
                                                             self.coord[1],
                                                             self.coord[2])

    def symbol(self) -> str:
        """

        :return: Returns the atomic symbol of the atom (from constants.py)
        """

        return number_to_symbol[self.charge]

    def xpos(self) -> float:
        """

        :return: Returns the x-coordinate of the atom
        """

        return self.coord[0]

    def ypos(self) -> float:
        """

        :return: Returns the y-coordinate of the atom
        """

        return self.coord[1]

    def zpos(self) -> float:
        """

        :return: Returns the z-coordinate of the atom
        """

        return self.coord[2]

    def set_charge(self, q: int):
        """
        This method re-sets the nuclear charge, i.e. the type of element.

        :param q: The nuclear charge (i.e. atomic number)
        :return: None
        """
        if q in number_to_symbol:
            self.charge = q

    def set_x(self, x: float) -> None:
        """
        This method re-sets the x-coordinate of the atom.

        :param x: Sets the cartesion x-coordinate of the atom to x.
        :return: None
        """
        self.coord[0] = x

    def set_y(self, y: float) -> None:
        """
        This method re-sets the y-coordinate of the atom.

        :param y: Sets the cartesion x-coordinate of the atom to y.
        :return: None
        """
        self.coord[1] = y

    def set_z(self, z: float) -> None:
        """
        This method re-sets the z-coordinate of the atom.

        :param z: Sets the cartesion x-coordinate of the atom to z.
        :return: None
        """
        self.coord[2] = z

    def set_xyz(self, x: float, y: float, z: float) -> None:
        """
        This method re-sets all x, y and z coordinates of the atom.

        :param x: Sets the cartesion x-coordinate of the atom to x.
        :param y: Sets the cartesion y-coordinate of the atom to y.
        :param z: Sets the cartesion x-coordinate of the atom to z.
        :return: None
        """
        self.coord[0] = x
        self.coord[1] = y
        self.coord[2] = z


def main():
    # Create all atoms in the periodic table up to ₁₁₈Og.
    for i in symbol_to_number:
        print(Atom(sym=i))


if __name__ == '__main__':
    main()



########################################################
# Molecule class and class methods to be defined below #
########################################################

import math
import itertools
import scipy.special
import numpy as np
from typing import Optional

from wellfare.atom import Atom
from wellfare.constants import symbol_to_mass, symbol_to_covalent_radius
from wellfare.messages import msg_program_error
from wellfare.parallelisation import EmbarrassingParallelisation


class Molecule:
    """A molecule with a name, charge, multiplicity and a list of atoms"""

    def __init__(self, name: str, charge: Optional[int] = 0,
                 multiplicity: Optional[int] = 1) -> None:
        """
        Creates a named Molecule with a charge, a multiplicity and, initially,
        no atoms. The charge is to be understood as the difference between the
        sum of nuclear charges and the number of electrons. The multiplicity is
        automatically set to the lowest possible value (1 or 2). There are also
        lists of bonds, angles and dihedrals (which are empty upon creation).
        The cartesian coordinates of the molecule are inherited from the
        individual cartesian coordinates of the atoms that make up the
        molecule (and are stored in self.atoms).

        :param name: The name of the molecule.
        :param charge: The overall molecular charge.
        :param multiplicity: The spin multiplicity.
        """

        self.name = name
        self.charge = charge
        self.mult = multiplicity
        self.atoms = []  # Initially an empty list
        self.bonds = []  # Initially an empty list
        self.angles = []  # Initially an empty list
        self.dihedrals = []  # Initially an empty list
        self.H_QM = np.zeros((3,
                              3))  # Force constants, Array size arbitrary, just a placeholder for type
        self.H_mw = np.zeros((3,
                              3))  # Mass weighted force constants, Array size arbitrary, just a placeholder for type
        self.frequencies = []  # List of vibrational frequencies in cm ⁻¹
        self.sigmaRot = 1  # Rotational symmetry number
        self.qm_energy = 0.0  # electronic energy in Hartree
        self.ZPVE = 0.0  # Zero Point Vibrational Energy in Hartree
        self.ZPVEList = []  # List of individual vibrations' contributions to the ZPVE in Hartree
        self.thermVib = 0.0  # Finite temperature vibrational contribution in Hartree
        self.thermVibList = []  # List of contributions from individual vibrations in Hartree
        self.thermRot = 0.0  # Thermal contribution from rotations in Hartree
        self.thermTrans = 0.0  # Thermal contribution from translation in Hartree
        self.kT = 0.0  # Thermal contribution (from pV = kT) to Enthalpy in Hartree
        self.transS = 0.0  # Translational entropy in J mol⁻¹ K⁻¹
        self.elecS = 0.0  # Electronic entropy in J mol⁻¹ K⁻¹
        self.rotS = 0.0  # Rotational entropy in J mol⁻¹ K⁻¹
        self.vibS = 0.0  # Vibrational entropy in J mol⁻¹ K⁻¹
        self.VibSList = []  # List of individual vibrations' contributions to the entropy in J mol⁻¹ K⁻¹
        self.negTS = 0.0  # Thermal contribution (from -TS) to Gibbs energy in Hartree

    def __str__(self) -> str:
        """
        Return a string representation of this Molecule in this format:
        (NAME, CHARGE, MULT, (ATOM1, ATOM2, ...))

        :return:
        """

        res = ''
        for atom in self.atoms:
            res = res + str(atom) + ', '
        res = res[:-2]
        return '({0}, {1}, {2}, ({3}))'.format(self.name, self.charge,
                                               self.mult, res)

    def __repr__(self) -> str:
        """
        Return a string representation of this Molecule in this format:
        (NAME, CHARGE, MULT, (ATOM1, ATOM2, ...))

        :return:
        """

        res = ''
        for atom in self.atoms:
            res = res + str(atom) + ', '
        res = res[:-2]
        return '({0}, {1}, {2}, ({3}))'.format(self.name, self.charge,
                                               self.mult, res)


    def mass(self) -> float:
        """

        :return: Returns the molar mass as sum of atomic masses
        """

        mass = 0.0
        for atom in self.atoms:
            mass = mass + atom.mass

        return mass


    def num_atoms(self) -> int:
        """

        :return: Returns the number of atoms in the molecule
        """

        return int(len(self.atoms))


    def add_atom(self, a: Atom, verbosity: int = 0) -> None:
        """
        This method adds an atom to the molecule. The molecule's multiplicity
        will be reset to the smallest possible value (singlet or doublet).

        :param a: An Atom object to be added to the molecule.
        :param verbosity: Print output about the action or not (default: no).
        :return: None
        """

        self.atoms.append(a)
        nucchg = 0
        for i in self.atoms:
            nucchg = nucchg + i.charge
        if (nucchg - self.charge) % 2 != 0:
            self.mult = 2
        else:
            self.mult = 1
        if verbosity >= 1:
            print(" adding {:<3} {: 13.8f} {: 13.8f} {: 13.8f} to {}".format(
                a.symbol(), a.xpos(), a.ypos(), a.zpos(), self.name))


    def ch_atom(self, n: int, at: Atom) -> None:
        """
        Exchanges the nth atom of the Molecule with a new atom.

        :param n: number of the atom to be exchanged.
        :param at: Atom object that replaces the old atom.
        :return: None
        """

        self.atoms[n] = at


    def mov_atom(self, n: int, x: float, y: float, z: float) -> None:
        """
        Move the nth atom to cartesian coordinates x, y, z.
        Coordinates are always given in Ångströms.

        :param n: number of the atom to be moved.
        :param x: target x-coordinate.
        :param y: target y-coordinate.
        :param z: target z-coordinate.
        :return: None
        """

        self.atoms[n].set_x(x)
        self.atoms[n].set_y(y)
        self.atoms[n].set_z(z)


    def add_bond(self, a: int, b: int) -> None:
        """
        Adds a bond between atoms a and b to the list of bonds. The indices a
        and b will be reordered to make sure that a < b and if the new bond is
        already present in the list of bonds, no second one will be added.

        :param a: First atom connected by the new bond.
        :param b: Second atom connected by the new bond.
        :return: None
        """

        # Make sure a < b
        if a < b:
            c = a
            d = b
        else:
            c = b
            d = a

        # Check if the bond already exists
        exists = False
        for i in self.bonds:
            if i == [c, d]:
                exists = True

        # Append bond to list if doesn't exist and is plausible
        if exists is False and a >= 0 and b >= 0 and a <= len(
                self.atoms) and b <= len(self.atoms) and c != d:
            self.bonds.append([c, d])


    def del_bond(self, a: int, b: int) -> None:
        """
        Deletes the bond between atoms a and b from the list of bonds.

        :param a: First atom connected by the bond.
        :param b: Second atom connected by the bond.
        :return: None
        """

        # Make sure a < b
        if a < b:
            c = a
            d = b
        else:
            c = b
            d = a

        # Check if the bond actually exists
        exists = False
        for i in self.bonds:
            if i == [c, d]:
                exists = True

        # Remove if it does
        if exists is True:
            self.bonds.remove([c, d])


    def add_angle(self, a: int, b: int, c: int) -> None:
        """
        Adds an angle between atoms a, b and c to the list of angles.

        :param a: The first atom that defines the angle.
        :param b: The second (i.e. middle) atom that defines the angle.
        :param c: The third atom that defines the angle.
        :return: None
        """

        # Check if the angle already exists
        exists = False
        for i in self.angles:
            if i == [a, b, c]:
                exists = True

        # Append angle to list if doesn't exist and is plausible.
        if exists is False and a >= 0 and b >= 0 and c >= 0 and \
                a <= len(self.atoms) and b <= len(self.atoms) and \
                c <= len(self.atoms) and a != b and a != c and b != c:
            self.angles.append([a, b, c])


    def add_dihedral(self, a: int, b: int, c: int, d: int) -> None:
        """
        Adds a dihedral between atoms a, b, c and d to the list of dihedrals.

        :param a: First atom defining the dihedral. Ordinarily bound to atom b.
        :param b: Second atom defining the dihedral. Ordinarily bound to atoms
                   a and b.
        :param c: Third atom defining the dihedral. Ordinarily bound to atoms
                   b and d.
        :param d: Last atom defining the dihedral. Ordinarily bound to atom c.
        :return: None
        """

        # Check if the dihedral already exists
        exists = False
        for i in self.dihedrals:
            if i == [a, b, c, d]:
                exists = True

        # Append dihedral to list if doesn't exist and is plausible
        if exists is False and a >= 0 and b >= 0 and c >= 0 and d >= 0 and \
                a <= len(self.atoms) and b <= len(self.atoms) and \
                c <= len(self.atoms) and d <= len(self.atoms) and \
                a != b and a != c and a != d and b != c and b != d and c != d:
            self.dihedrals.append([a, b, c, d])


    def atm_symbol(self, i: int) -> str:
        """
        Report the atomic symbol of atom i.

        :param i: Atom to locate.
        :return: The atom's atomic symbol
        """
        return self.atoms[i].symbol()


    def atm_pos_x(self, i: int) -> float:
        """
        Report the x-coordinate of atom i in Ångströms.

        :param i: Atom to locate.
        :return: The atoms x-coordinate in Ångströms.
        """
        return self.atoms[i].coord[0]


    def atm_pos_y(self, i: int) -> float:
        """
        Report the y-coordinate of atom i in Ångströms.

        :param i: Atom to locate.
        :return: The atoms y-coordinate in Ångströms.
        """
        return self.atoms[i].coord[1]


    def atm_pos_z(self, i: int) -> float:
        """
        Report the z-coordinate of atom i in Ångströms.

        :param i: Atom to locate.
        :return: The atoms z-coordinate in Ångströms.
        """
        return self.atoms[i].coord[2]


    def atm_atm_dist(self, i: int, j: int) -> float:
        """
        Report the distance between atoms i and j in Ångström.

        :param i: First atom for distance measurement.
        :param j: Second atom for distance measurement.
        :return: The distance between atoms i and j in Ångström.
        """

        distance = (self.atm_pos_x(i) - self.atm_pos_x(j)) * (
                self.atm_pos_x(i) - self.atm_pos_x(j))
        distance = distance + (self.atm_pos_y(i) - self.atm_pos_y(j)) * (
                self.atm_pos_y(i) - self.atm_pos_y(j))
        distance = distance + (self.atm_pos_z(i) - self.atm_pos_z(j)) * (
                self.atm_pos_z(i) - self.atm_pos_z(j))

        return math.sqrt(distance)


    def bond_dist(self, i: int) -> float:
        """
        Report the distance between two atoms in the bonds list in Ångström.

        :param i: Index of the bond from the list of bonds..
        :return: The distance between the in Ångström.
        """

        atom1 = self.bonds[i][0]
        atom2 = self.bonds[i][1]
        distance = self.atm_atm_dist(atom1, atom2)

        return distance


    def atm_atm_atm_angle(self, i: int, j: int, k: int) -> float:
        """
        Report the angle between three atoms in radians.

        :param i: First atom for angle measurement.
        :param j: Second atom for angle measurement.
        :param k: Third atom for angle measurement.
        :return: The numerical value of the angle in radians.
        """

        # Calculate the distance between each pair of atoms
        d_bond_1 = self.atm_atm_dist(i, j)
        d_bond_2 = self.atm_atm_dist(j, k)
        d_non_bond = self.atm_atm_dist(i, k)

        # Use those distances and the cosine rule to calculate bond angle theta
        numerator = d_bond_1 ** 2 + d_bond_2 ** 2 - d_non_bond ** 2
        denominator = 2 * d_bond_1 * d_bond_2
        argument = numerator / denominator
        # This safety check was necessary because of a bug elsewhere once...
        # if argument > 1.0:
        #     argument = 1.0
        # elif argument < -1.0:
        #     argument = -1.0
        theta = np.arccos(argument)

        return theta


    def bond_angle(self, i: int) -> float:
        """
        Report the angle described by three atoms in the bonds list in radians.

        :param i: Index of the bond angle from the list of angles.
        :return: The numerical value of the angle in radians.
        """
        atom1 = self.angles[i][0]
        atom2 = self.angles[i][1]
        atom3 = self.angles[i][2]
        angle = self.atm_atm_atm_angle(atom1, atom2, atom3)

        return angle


    def atm_atm_atm_atm_dihedral(self, i: int, j: int, k: int,
                                 l: int) -> float:
        """
        Report the dihedral angle described by a set of four atoms in radians.


        :param i: First atom for angle measurement.
        :param j: Second atom for angle measurement.
        :param k: Third atom for angle measurement.
        :param l: Fourth atom for angle measurement.
        :return: The numerical value of the dihedral angle in radians.
        """

        # Calculate the vectors lying along bonds, and their cross products
        atom_e1 = self.atoms[i]
        atom_b1 = self.atoms[j]
        atom_b2 = self.atoms[k]
        atom_e2 = self.atoms[l]
        end_1 = [atom_e1.coord[idx] - atom_b1.coord[idx] for idx in range(3)]
        bridge = [atom_b1.coord[idx] - atom_b2.coord[idx] for idx in range(3)]
        end_2 = [atom_b2.coord[idx] - atom_e2.coord[idx] for idx in range(3)]
        vnormal_1 = np.cross(end_1, bridge)
        vnormal_2 = np.cross(bridge, end_2)

        # Construct a set of orthogonal basis vectors to define a frame with
        # vnormal_2 as the x axis
        vcross = np.cross(vnormal_2, bridge)
        norm_vn2 = np.linalg.norm(vnormal_2)
        # norm_b = np.linalg.norm(bridge)
        norm_vc = np.linalg.norm(vcross)
        basis_vn2 = [vnormal_2[idx] / norm_vn2 for idx in range(3)]
        # basis_b = [bridge[idx] / norm_b for idx in range(3)]
        basis_cv = [vcross[idx] / norm_vc for idx in range(3)]

        # Find the signed angle betw. vnormal_1 and vnormal_2 in the new frame
        vn1_coord_n2 = np.dot(vnormal_1, basis_vn2)
        vn1_coord_vc = np.dot(vnormal_1, basis_cv)
        psi = math.atan2(vn1_coord_vc, vn1_coord_n2)

        return psi

    def dihedral_angle(self, i: int) -> float:
        """
        Report the dihedral angle described by a set of four atoms in the list
        of dihedrals in radians.

        :param i: Index of the dihedral angle in the list of dihedrals.
        :return: The numerical value of the angle in radians.
        """

        atom1 = self.dihedrals[i][0]
        atom2 = self.dihedrals[i][1]
        atom3 = self.dihedrals[i][2]
        atom4 = self.dihedrals[i][3]

        dihedral = self.atm_atm_atm_atm_dihedral(atom1, atom2, atom3, atom4)

        return dihedral


    def set_hessian(self, H):
        """ (Molecule) -> NoneType

        Set the Quantum Mechanically calculated Hessian, H_QM, equal to H
        """
        self.H_QM = H


    def set_mult(self, multiplicity: int) -> None:
        """
        Set the multiplicity of the molecule to M

        :param multiplicity: Value of the multiplicity (i.e. 1 for singlet,
                             2 for doublet, ...)
        :return: None
        """

        self.mult = multiplicity


    def set_charge(self, charge: int) -> None:
        """
        Set the charge of the molecule to C

        :param charge: Value of the charge (i.e. -1 for a singly charged anion,
                                           2 for a doubly charged cation...)
        :return: None
        """

        self.charge = charge


    def print_mol(self, output: str = "cart", comment: Optional[str] = None,
                  file: Optional[str] = None) -> str:
        """
        Returns a string containing the molecular coordinates, which can also
        be written to a file.

        If no input at all is given, we use the "name" of the molecule as the
        comment line. If a file name is given, but no comment, we use the
        file name as comment line. If a comment is specified, we use that as
        the comment line (Note: if an empty string is desired, then "" must
        be submitted as comment)

        :param output: Style of the output:
                         "cart" for plain cartesian coordinates.
                         "xyz" for the xyz-file format (i.e. with # of atoms).
                         "gauss" for Gaussian style.
                         "gamess" for Gamess style.
        :param comment: A comment (string) that can be printed into the output
        :param file: The filename of a file that is created and written to.
        :return: The string with the molecular coordinates.
        """

        if comment is None and file is None:
            commentline = self.name
        elif comment is None and file is not None:
            commentline = file
        else:
            commentline = comment

        s = ""
        if output == "cart":
            # Plain cartesian coordinates - nothing else
            for i in self.atoms:
                t = "{:<3} {: 13.8f} {: 13.8f} {: 13.8f}\n".format(i.symbol(),
                                                                   i.xpos(),
                                                                   i.ypos(),
                                                                   i.zpos())
                s = s + t
        elif output == "xyz":
            # xyz-file format. Number of atoms first, comment line,
            # then coordinates
            s = str(self.num_atoms()) + "\n" + commentline + "\n"
            for i in self.atoms:
                t = "{:<3} {: 13.8f} {: 13.8f} {: 13.8f}\n".format(i.symbol(),
                                                                   i.xpos(),
                                                                   i.ypos(),
                                                                   i.zpos())
                s = s + t
        elif output == "gamess":
            # xyz coordinates in Gamess format
            s = " $DATA\n" + commentline + "\nC1\n"
            for i in self.atoms:
                t = "{:<3} {:<3d} {: 13.8f} {: 13.8f} {: 13.8f}\n".format(
                    i.symbol(),
                    i.charge,
                    i.xpos(),
                    i.ypos(),
                    i.zpos())
                s = s + t
            s = s + " $END\n"
        elif output == "gauss":
            # xyz coordinates in Gaussian format
            s = "\n" + commentline + "\n\n" + str(self.charge) + " " + str(
                self.mult) + "\n"
            for i in self.atoms:
                if i.mass == symbol_to_mass[i.symbol()]:
                    t = "{:<3} {: 13.8f} {: 13.8f} {: 13.8f}\n".format(
                        i.symbol(),
                        i.xpos(),
                        i.ypos(),
                        i.zpos())
                else:
                    t = "{} {: 13.8f} {: 13.8f} {: 13.8f}\n".format(
                        i.symbol() + "(Iso=" + str(i.mass) + ")",
                        i.xpos(),
                        i.ypos(),
                        i.zpos())
                s = s + t
            s = s + "\n"

        if file is not None:
            # Try filename for writability first
            try:
                f = open(file + "." + output, 'w')
                f.close()
            except OSError:
                msg_program_error("Cannot open " + file + " for writing.")
            # Then write the string we created above
            with open(file + "." + output, "w") as text_file:
                print("{}".format(s), file=text_file)

        return s


def batch_identify_dihedrals(molecule, chunk_start, chunk_end, verbosity=0):
    results = []
    iterator = itertools.combinations(range(len(molecule.angles)), 2)
    for i, j in itertools.islice(iterator, chunk_start, chunk_end):
        if molecule.angles[i][1] == molecule.angles[j][0] and \
                molecule.angles[i][2] == molecule.angles[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_atm_dihedral(
                    molecule.angles[i][0], molecule.angles[i][1],
                    molecule.angles[j][1],
                    molecule.angles[j][2]))
            results.append(
                [molecule.angles[i][0], molecule.angles[i][1],
                 molecule.angles[j][1], molecule.angles[j][2], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({: 7.2f}°)".format(
                        molecule.atm_symbol(molecule.angles[i][0]),
                        molecule.angles[i][0],
                        molecule.atm_symbol(molecule.angles[i][1]),
                        molecule.angles[i][1],
                        molecule.atm_symbol(molecule.angles[j][1]),
                        molecule.angles[j][1],
                        molecule.atm_symbol(molecule.angles[j][2]),
                        molecule.angles[j][2], angle))
        if molecule.angles[i][1] == molecule.angles[j][2] and \
                molecule.angles[i][2] == molecule.angles[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_atm_dihedral(
                    molecule.angles[i][0], molecule.angles[i][1],
                    molecule.angles[j][1],
                    molecule.angles[j][0]))
            results.append(
                [molecule.angles[i][0], molecule.angles[i][1],
                 molecule.angles[j][1], molecule.angles[j][0], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({: 7.2f}°)".format(
                        molecule.atm_symbol(molecule.angles[i][0]),
                        molecule.angles[i][0],
                        molecule.atm_symbol(molecule.angles[i][1]),
                        molecule.angles[i][1],
                        molecule.atm_symbol(molecule.angles[j][1]),
                        molecule.angles[j][1],
                        molecule.atm_symbol(molecule.angles[j][0]),
                        molecule.angles[j][0], angle))
        if molecule.angles[i][1] == molecule.angles[j][0] and \
                molecule.angles[i][0] == molecule.angles[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_atm_dihedral(
                    molecule.angles[i][2], molecule.angles[i][1],
                    molecule.angles[j][1],
                    molecule.angles[j][2]))
            results.append(
                [molecule.angles[i][2], molecule.angles[i][1],
                 molecule.angles[j][1], molecule.angles[j][1], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({: 7.2f}°)".format(
                        molecule.atm_symbol(molecule.angles[i][2]),
                        molecule.angles[i][2],
                        molecule.atm_symbol(molecule.angles[i][1]),
                        molecule.angles[i][1],
                        molecule.atm_symbol(molecule.angles[j][1]),
                        molecule.angles[j][1],
                        molecule.atm_symbol(molecule.angles[j][2]),
                        molecule.angles[j][2], angle))
        if molecule.angles[i][1] == molecule.angles[j][2] and \
                molecule.angles[i][0] == molecule.angles[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_atm_dihedral(
                    molecule.angles[i][2], molecule.angles[i][1],
                    molecule.angles[j][1],
                    molecule.angles[j][0]))
            results.append(
                [molecule.angles[i][2], molecule.angles[i][1],
                 molecule.angles[j][1], molecule.angles[j][0], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({: 7.2f}°)".format(
                        molecule.atm_symbol(molecule.angles[i][2]),
                        molecule.angles[i][2],
                        molecule.atm_symbol(molecule.angles[i][1]),
                        molecule.angles[i][1],
                        molecule.atm_symbol(molecule.angles[j][1]),
                        molecule.angles[j][1],
                        molecule.atm_symbol(molecule.angles[j][0]),
                        molecule.angles[j][0], angle))
    return results


def build_molecular_dihedrals(molecule, verbosity=0, deletefirst=True,
                              canonical_order=False, cpu_number=1):
    """

    :param molecule: The molecule to be operated on
    :param verbosity: How chatty this method is
    :param deletefirst: Should the list of dihedrals be deleted, before a
                        new list is built?
    :param cpu_number: The number of processor cores to be used via the
                       multiprocessing module.
    :param canonical_order: Setting this to True will make sure the list of
                            dihedrals (and the list of returned values) are
                            sorted (the list might be returned in arbitrary
                            order otherwise because of the asynchronous
                            parallelisation).
    :return: A list of the values of the dihedral angles in the same order as
             the dihedrals that have been identified
    """

    list_of_dihedrals = []

    if deletefirst is True:
        molecule.dihedrals = []

    if len(molecule.angles) > 1:
        if verbosity >= 2:
            print("\nLooking for dihedrals in WellFARe molecule: ",
                  molecule.name)
        if verbosity >= 2 and cpu_number > 1:
            print(
                "\nStarting parallel execution on {} processor"
                " cores.".format(cpu_number))
        number_of_pairs = int(scipy.special.comb(N=len(molecule.angles), k=2))
        with EmbarrassingParallelisation(number_of_processes=cpu_number,
                                         number_of_ops=number_of_pairs) as pe:
            res = []
            for chunk in pe.chunks:
                res.append(pe.pool.apply_async(batch_identify_dihedrals, args=(
                    molecule, chunk.start, chunk.finish, verbosity)))
            results = [p.get() for p in res]
        for i in results:
            for j in i:
                molecule.add_dihedral(j[0], j[1], j[2], j[3])
                list_of_dihedrals.append(j[4])
        if canonical_order is True and len(list_of_dihedrals) > 0:
            # If called for, we sort the list of angles (might be in arbitrary
            # order because of asynchronous parallelism "apply_async" above).
            zipped_list = sorted(
                list(zip(molecule.angles, list_of_dihedrals)))
            molecule.angles, list_of_dihedrals = list(zip(*zipped_list))

    else:
        if verbosity >= 2:
            print("\nNot looking for dihedrals in WellFARe molecule: ",
                  molecule.name)
            print(
                "(because there are fewer than two angles"
                " identified in the molecule)")
    return list_of_dihedrals


def batch_identify_angles(molecule, chunk_start, chunk_end, verbosity=0):
    results = []
    iterator = itertools.combinations(range(len(molecule.bonds)), 2)
    for i, j in itertools.islice(iterator, chunk_start, chunk_end):
        if molecule.bonds[i][0] == molecule.bonds[j][0]:
            angle = math.degrees(
                molecule.atm_atm_atm_angle(molecule.bonds[i][1],
                                           molecule.bonds[i][0],
                                           molecule.bonds[j][1]))
            results.append([molecule.bonds[i][1], molecule.bonds[i][0],
                            molecule.bonds[j][1], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({:6.2f}°)".format(
                        molecule.atm_symbol(molecule.bonds[i][1]),
                        molecule.bonds[i][1],
                        molecule.atm_symbol(molecule.bonds[i][0]),
                        molecule.bonds[i][0],
                        molecule.atm_symbol(molecule.bonds[j][1]),
                        molecule.bonds[j][1], angle))
        if molecule.bonds[i][0] == molecule.bonds[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_angle(molecule.bonds[i][1],
                                           molecule.bonds[i][0],
                                           molecule.bonds[j][0]))
            results.append([molecule.bonds[i][1], molecule.bonds[i][0],
                            molecule.bonds[j][0], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({:6.2f}°)".format(
                        molecule.atm_symbol(molecule.bonds[i][1]),
                        molecule.bonds[i][1],
                        molecule.atm_symbol(molecule.bonds[i][0]),
                        molecule.bonds[i][0],
                        molecule.atm_symbol(molecule.bonds[j][0]),
                        molecule.bonds[j][0], angle))
        if molecule.bonds[i][1] == molecule.bonds[j][0]:
            angle = math.degrees(
                molecule.atm_atm_atm_angle(molecule.bonds[i][0],
                                           molecule.bonds[j][0],
                                           molecule.bonds[j][1]))
            results.append([molecule.bonds[i][0], molecule.bonds[j][0],
                            molecule.bonds[j][1], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({:6.2f}°)".format(
                        molecule.atm_symbol(molecule.bonds[i][0]),
                        molecule.bonds[i][0],
                        molecule.atm_symbol(molecule.bonds[j][0]),
                        molecule.bonds[j][0],
                        molecule.atm_symbol(molecule.bonds[j][1]),
                        molecule.bonds[j][1], angle))
        if molecule.bonds[i][1] == molecule.bonds[j][1]:
            angle = math.degrees(
                molecule.atm_atm_atm_angle(molecule.bonds[i][0],
                                           molecule.bonds[j][1],
                                           molecule.bonds[j][0]))
            results.append([molecule.bonds[i][0], molecule.bonds[j][1],
                            molecule.bonds[j][0], angle])
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}), {:<3} ({:3d})"
                    " and {:<3} ({:3d}) ({:6.2f}°)".format(
                        molecule.atm_symbol(molecule.bonds[i][0]),
                        molecule.bonds[i][0],
                        molecule.atm_symbol(molecule.bonds[j][1]),
                        molecule.bonds[j][1],
                        molecule.atm_symbol(molecule.bonds[j][0]),
                        molecule.bonds[j][0], angle))
    return results


def build_molecular_angles(molecule, verbosity=0, deletefirst=True,
                           canonical_order=False, cpu_number=1):
    """

    :param molecule: The molecule to be operated on
    :param verbosity: How chatty this method is.
    :param deletefirst: Should the list of angles be deleted, before a new list
                        is built?
    :param cpu_number: The number of processor cores to be used via the
                       multiprocessing module.
    :param canonical_order: Setting this to True will make sure the list of
                            angles (and the list of returned degree values) are
                            sorted (the list might be returned in arbitrary
                            order otherwise because of the asynchronous
                            parallelisation).
    :return: A list with the values of all bond angles in degrees in the same
             order as the bond angles.
    """

    if deletefirst is True:
        molecule.angles = []

    list_of_bond_angles = []

    # Now that we know where the bonds are, find angles
    if len(molecule.bonds) > 1:
        if verbosity >= 2:
            print("\nLooking for angles in WellFARe molecule: ", molecule.name)
        if verbosity >= 2 and cpu_number > 1:
            print(
                "\nStarting parallel execution on {} processor"
                " cores.".format(cpu_number))
        number_of_pairs = int(scipy.special.comb(N=len(molecule.bonds), k=2))
        with EmbarrassingParallelisation(number_of_processes=cpu_number,
                                         number_of_ops=number_of_pairs) as pe:
            res = []
            for chunk in pe.chunks:
                res.append(pe.pool.apply_async(batch_identify_angles,
                                               args=(molecule, chunk.start,
                                                     chunk.finish, verbosity)))
            results = [p.get() for p in res]
        for i in results:
            for j in i:
                molecule.add_angle(j[0], j[1], j[2])
                list_of_bond_angles.append(j[3])
        if canonical_order is True and len(list_of_bond_angles) > 0:
            # If called for, we sort the list of angles (might be in arbitrary
            # order because of asynchronous parallelism "apply_async" above).
            zipped_list = sorted(
                list(zip(molecule.angles, list_of_bond_angles)))
            molecule.angles, list_of_bond_angles = list(zip(*zipped_list))
    else:
        if verbosity >= 2:
            print("\nNot looking for angles in WellFARe"
                  " molecule: {}".format(molecule.name))
            print("(because there are fewer than 2 bonds"
                  " identified in the molecule)")
    return list_of_bond_angles


def batch_compare_distances(molecule, chunk_start, chunk_end, tolerance,
                            verbosity=0):
    results = []
    iterator = itertools.combinations(range(molecule.num_atoms()), 2)
    for i, j in itertools.islice(iterator, chunk_start, chunk_end):
        distance = molecule.atm_atm_dist(i, j)
        if distance <= (
                symbol_to_covalent_radius[molecule.atm_symbol(i)] +
                symbol_to_covalent_radius[
                    molecule.atm_symbol(j)]) * tolerance:
            if verbosity >= 3:
                print(
                    " {:<3} ({:3d}) and {:<3} ({:3d})"
                    " (Distance: {:.3f} Å)".format(
                        molecule.atm_symbol(i), i,
                        molecule.atm_symbol(j), j,
                        distance))
            results.append([i, j, distance])
    return results


def build_bond_orders(molecule, bo=None, verbosity=0, bondcutoff=0.45,
                      distfactor=1.3, deletefirst=True, canonical_order=False,
                      cpu_number=1):
    """
    This method populates the list of bonds of a given molecule.

    :param molecule: The molecule to operate on
    :param bo: an n by n matrix (i.e. array) that holds qm bond orders (if not
               available, bonds will be built from distances alone).
    :param verbosity: How talkative this method is, default is silence.
    :param bondcutoff: Cutoff value for the qm bond orders below which no bond
                       will be assigned between two atoms.
    :param distfactor: Cutoff value for a multiple of the sum of covalent radii
                       above which no bond will be assigned between two atoms.
    :param deletefirst: Should the list of bonds be deleted, before a new list
                        is built?
    :param cpu_number: The number of processor cores to be used via the
                       multiprocessing module.
    :param canonical_order: Setting this to True will make sure the list of
                            bonds (and the list of returned bond lengths) are
                            sorted (the list might be returned in arbitrary
                            order otherwise because of the asynchronous
                            parallelisation).
    :return: A list of all bond lengths in Å in the same order as the bonds.
    """

    if bo is None:
        bo = []
    list_of_bond_lengths = []

    if deletefirst is True:
        molecule.bonds = []

    # Test if we actually have Quantum Mechanical (qm) Bond orders
    if np.count_nonzero(bo) != 0 and molecule.num_atoms() != 1:
        if verbosity >= 2:
            print("\nLooking for bonds in WellFARe molecule: ",
                  molecule.name)
            print("(using bond orders with a cutoff of {: .2f}):".format(
                bondcutoff))
        for i in range(0, molecule.num_atoms()):
            for j in range(i + 1, molecule.num_atoms()):
                if bo[i][j] >= bondcutoff:
                    molecule.add_bond(i, j)
                    list_of_bond_lengths.append(
                        molecule.atm_atm_dist(i, j))
                    if verbosity >= 3:
                        print(
                            " {:<3} ({:4d}) and {:<3} ({:4d})"
                            " (Bond order: {: 6.3f}, length {:6.3f} Å)".format(
                                molecule.atoms[i].symbol(), i,
                                molecule.atoms[j].symbol(), j,
                                float(bo[i][j]), list_of_bond_lengths[-1]))
        if canonical_order is True and len(list_of_bond_lengths) > 0:
            # If called for, we sort the list of bonds (see "elif" below)
            # Unlikely to be necessary in the case of qm bond-orders...
            zipped_list = sorted(
                list(zip(molecule.bonds, list_of_bond_lengths)))
            molecule.bonds, list_of_bond_lengths = list(zip(*zipped_list))
    # Only do bonds if there's more than one atom
    elif molecule.num_atoms() != 1:
        if verbosity >= 2:
            print("\nLooking for bonds in WellFARe molecule:", molecule.name)
            print(
                "(using covalent radii scaled by {: .2f}):".format(distfactor))
        if verbosity >= 2 and cpu_number > 1:
            print(
                "\nStarting parallel execution on {} processor"
                " cores.".format(cpu_number))
        number_of_pairs = int(scipy.special.comb(N=molecule.num_atoms(), k=2))
        with EmbarrassingParallelisation(number_of_processes=cpu_number,
                                         number_of_ops=number_of_pairs) as pe:
            res = []
            for chunk in pe.chunks:
                res.append(pe.pool.apply_async(batch_compare_distances,
                                               args=(molecule, chunk.start,
                                                     chunk.finish,
                                                     distfactor, verbosity)))
            results = [p.get() for p in res]
        for i in results:
            for j in i:
                molecule.add_bond(j[0], j[1])
                list_of_bond_lengths.append(j[2])
        if canonical_order is True and len(list_of_bond_lengths) > 0:
            # If called for, we sort the list of bonds (might be in arbitrary
            # order because of asynchronous parallelism "apply_async" above).
            zipped_list = sorted(
                list(zip(molecule.bonds, list_of_bond_lengths)))
            molecule.bonds, list_of_bond_lengths = list(zip(*zipped_list))
    else:
        if verbosity >= 2:
            print("\nNot looking for bonds in WellFARe molecule:",
                  molecule.name)
            print("(because we only have one atom)")
    return list_of_bond_lengths


def main():
    # Create an example molecule and add some atoms
    example = Molecule("Example Molecule")
    example.add_atom(Atom("C", x=-0.63397128, y=0.57416267, z=0.000000))
    example.add_atom(Atom("H", x=-0.27731685, y=-0.43464733, z=0.000000))
    example.add_atom(Atom("H", x=-0.27729844, y=1.07856086, z=0.87365150))
    example.add_atom(Atom("H", x=-0.27729844, y=1.07856086, z=-0.87365150))
    example.add_atom(
        Atom("H", x=-1.70397128, y=0.57417585, z=0.00000000, mass=2.0))
    print(example.print_mol(output="gauss"))


if __name__ == '__main__':
    main()



from setuptools import setup

setup(
    name='wellfare',
    version='1.0.3',
    python_requires='>=3.4',
    install_requires=['scipy', 'numpy', 'matplotlib', 'typing'],
    packages=['wellfare'],
    scripts={'wellfare/comparestruct', 'wellfare/freeda', 'wellfare/mincepies',
             'wellfare/specplot', 'wellfare/wellfare'},
    url='https://github.com/matthiaslein/WellFARe',
    license='GNU General Public License v3.0',
    author='Matthias Lein',
    author_email='matthias.lein@gmail.com',
    description='WellFARe - The Wellington Fast Assessment of REactions',
    long_description='The wellfare package is a collection of quantum chemical'
                     ' programs that interact with electronic structure'
                     ' packages in order to aid analysis and computation. The'
                     ' distribution includes a tool to compare 3d structures,'
                     ' the Free Energy Decomposition Analysis, the mincepies'
                     ' tool for conformer generation and the wellfare force'
                     ' field itself.'
)
